"""Test for played move injection feature."""

import pytest

from ankigammon.models import Position, Player, Decision, Move, DecisionType
from ankigammon.settings import Settings


class TestPlayedMoveInjection:
    """Test that played moves are injected into top N candidates for MCQ display."""

    def setup_method(self):
        """Set up test fixtures."""
        self.settings = Settings()
        self.settings.max_mcq_options = 5  # Default to 5 for these tests

    def test_played_move_already_in_top_5(self):
        """Test that nothing happens if played move is already in top N."""
        position = Position()

        # Create 7 moves where played move is 3rd
        moves = [
            Move(notation="13/9 6/5", equity=0.234, rank=1, was_played=False),
            Move(notation="24/20 13/9", equity=0.221, error=0.013, rank=2, was_played=False),
            Move(notation="13/7 6/3", equity=0.210, error=0.024, rank=3, was_played=True),  # Played
            Move(notation="24/18 13/10", equity=0.200, error=0.034, rank=4, was_played=False),
            Move(notation="13/10 13/7", equity=0.190, error=0.044, rank=5, was_played=False),
            Move(notation="24/21 24/18", equity=0.180, error=0.054, rank=6, was_played=False),
            Move(notation="6/3 6/off", equity=0.170, error=0.064, rank=7, was_played=False),
        ]

        decision = Decision(
            position=position,
            on_roll=Player.O,
            dice=(6, 3),
            candidate_moves=moves,
            decision_type=DecisionType.CHECKER_PLAY
        )

        played_move = moves[2]  # 3rd move

        # Create a minimal MainWindow mock
        class MainWindowMock:
            def __init__(self, settings):
                self.settings = settings

            def _ensure_played_move_in_candidates(self, decision, played_move):
                # Copy the implementation with configurable max_options
                max_options = self.settings.max_mcq_options
                top_n = decision.candidate_moves[:max_options]
                if played_move in top_n:
                    return
                decision.candidate_moves.remove(played_move)
                decision.candidate_moves.insert(max_options - 1, played_move)

        mock = MainWindowMock(self.settings)
        mock._ensure_played_move_in_candidates(decision, played_move)

        # Verify: candidate_moves should be unchanged
        assert decision.candidate_moves[2] == played_move
        assert len(decision.candidate_moves) == 7

    def test_played_move_not_in_top_5(self):
        """Test that played move is injected when it's not in top N."""
        position = Position()

        # Create 7 moves where played move is 6th (a blunder)
        moves = [
            Move(notation="13/9 6/5", equity=0.234, rank=1, was_played=False),
            Move(notation="24/20 13/9", equity=0.221, error=0.013, rank=2, was_played=False),
            Move(notation="13/7 6/3", equity=0.210, error=0.024, rank=3, was_played=False),
            Move(notation="24/18 13/10", equity=0.200, error=0.034, rank=4, was_played=False),
            Move(notation="13/10 13/7", equity=0.190, error=0.044, rank=5, was_played=False),
            Move(notation="24/21 24/18", equity=0.100, error=0.134, rank=6, was_played=True),  # Played - big blunder!
            Move(notation="6/3 6/off", equity=0.090, error=0.144, rank=7, was_played=False),
        ]

        decision = Decision(
            position=position,
            on_roll=Player.O,
            dice=(6, 3),
            candidate_moves=moves,
            decision_type=DecisionType.CHECKER_PLAY
        )

        played_move = moves[5]  # 6th move (blunder)
        original_5th = moves[4]  # Save reference BEFORE injection

        # Create a minimal MainWindow mock
        class MainWindowMock:
            def __init__(self, settings):
                self.settings = settings

            def _ensure_played_move_in_candidates(self, decision, played_move):
                # Copy the implementation with configurable max_options
                max_options = self.settings.max_mcq_options
                top_n = decision.candidate_moves[:max_options]
                if played_move in top_n:
                    return
                decision.candidate_moves.remove(played_move)
                decision.candidate_moves.insert(max_options - 1, played_move)

        mock = MainWindowMock(self.settings)
        mock._ensure_played_move_in_candidates(decision, played_move)

        # Verify: played move should now be at position 4 (5th slot, with max_mcq_options=5)
        max_options = self.settings.max_mcq_options
        assert decision.candidate_moves[max_options - 1] == played_move
        assert decision.candidate_moves[max_options - 1].was_played is True
        assert decision.candidate_moves[max_options - 1].notation == "24/21 24/18"

        # The Nth best move should have been pushed down
        top_n_after = decision.candidate_moves[:max_options]
        assert played_move in top_n_after

        # Original Nth move should now be at position N or later
        assert decision.candidate_moves[max_options - 1] != original_5th
        assert decision.candidate_moves[max_options - 1].notation == "24/21 24/18"  # Played move
        assert original_5th.notation == "13/10 13/7"  # Original 5th move

    def test_played_move_is_last(self):
        """Test that played move is injected when it's the worst move."""
        position = Position()

        # Create 10 moves where played move is last (terrible blunder)
        moves = [
            Move(notation=f"Move{i}", equity=1.0 - i*0.1, error=i*0.1, rank=i+1, was_played=(i==9))
            for i in range(10)
        ]

        decision = Decision(
            position=position,
            on_roll=Player.O,
            dice=(6, 3),
            candidate_moves=moves,
            decision_type=DecisionType.CHECKER_PLAY
        )

        played_move = moves[9]  # Last move (worst blunder)

        # Create a minimal MainWindow mock
        class MainWindowMock:
            def __init__(self, settings):
                self.settings = settings

            def _ensure_played_move_in_candidates(self, decision, played_move):
                # Copy the implementation with configurable max_options
                max_options = self.settings.max_mcq_options
                top_n = decision.candidate_moves[:max_options]
                if played_move in top_n:
                    return
                decision.candidate_moves.remove(played_move)
                decision.candidate_moves.insert(max_options - 1, played_move)

        mock = MainWindowMock(self.settings)
        mock._ensure_played_move_in_candidates(decision, played_move)

        # Verify: played move should now be at position N-1 (last slot)
        max_options = self.settings.max_mcq_options
        assert decision.candidate_moves[max_options - 1] == played_move
        assert decision.candidate_moves[max_options - 1].was_played is True

        # Verify it's now in top N
        top_n_after = decision.candidate_moves[:max_options]
        assert played_move in top_n_after

    def test_played_move_with_custom_max_options(self):
        """Test that function respects custom max_mcq_options setting."""
        # Test with max_mcq_options = 3
        self.settings.max_mcq_options = 3
        position = Position()

        # Create 6 moves where played move is 5th (outside top 3)
        moves = [
            Move(notation="13/9 6/5", equity=0.234, rank=1, was_played=False),
            Move(notation="24/20 13/9", equity=0.221, error=0.013, rank=2, was_played=False),
            Move(notation="13/7 6/3", equity=0.210, error=0.024, rank=3, was_played=False),
            Move(notation="24/18 13/10", equity=0.200, error=0.034, rank=4, was_played=False),
            Move(notation="24/21 24/18", equity=0.100, error=0.134, rank=5, was_played=True),  # Played - outside top 3
            Move(notation="6/3 6/off", equity=0.090, error=0.144, rank=6, was_played=False),
        ]

        decision = Decision(
            position=position,
            on_roll=Player.O,
            dice=(6, 3),
            candidate_moves=moves,
            decision_type=DecisionType.CHECKER_PLAY
        )

        played_move = moves[4]  # 5th move (outside top 3)

        # Create a minimal MainWindow mock
        class MainWindowMock:
            def __init__(self, settings):
                self.settings = settings

            def _ensure_played_move_in_candidates(self, decision, played_move):
                # Copy the implementation with configurable max_options
                max_options = self.settings.max_mcq_options
                top_n = decision.candidate_moves[:max_options]
                if played_move in top_n:
                    return
                decision.candidate_moves.remove(played_move)
                decision.candidate_moves.insert(max_options - 1, played_move)

        mock = MainWindowMock(self.settings)
        mock._ensure_played_move_in_candidates(decision, played_move)

        # Verify: played move should now be at position 2 (3rd slot, with max_mcq_options=3)
        assert decision.candidate_moves[2] == played_move
        assert decision.candidate_moves[2].was_played is True

        # Verify it's now in top 3
        top_3_after = decision.candidate_moves[:3]
        assert played_move in top_3_after
        assert len(top_3_after) == 3



#
#   xgzarc.py - XG ZLib archive module
#   Copyright (C) 2013,2014  Michael Petch <mpetch@gnubg.org>
#                                          <mpetch@capp-sysware.com>
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Lesser General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#
#   This library is an interpretation of ZLBArchive 1.52 data structures.
#   Please see: http://www.delphipages.com/comp/zlibarchive-2104.html
#

from __future__ import with_statement as _with
import tempfile as _tempfile
import struct as _struct
import zlib as _zlib
import os as _os
from . import xgutils as _xgutils


class Error(Exception):

    def __init__(self, error):
        self.value = "Zlib archive: %s" % str(error)
        self.error = error

    def __str__(self):
        return repr(self.value)


class ArchiveRecord(dict):

    SIZEOFREC = 36

    def __init__(self, **kw):
        defaults = {
            'crc': 0,
            'filecount': 0,
            'version': 0,
            'registrysize': 0,
            'archivesize': 0,
            'compressedregistry': False,
            'reserved': []
            }
        super(ArchiveRecord, self).__init__(defaults, **kw)

    def __setattr__(self, key, value):
        self[key] = value

    def __getattr__(self, key):
        return self[key]

    def fromstream(self, stream):
        unpacked_data = _struct.unpack('<llllll12B',
                                       stream.read(self.SIZEOFREC))
        self.crc = unpacked_data[0] & 0xffffffff
        self.filecount = unpacked_data[1]
        self.version = unpacked_data[2]
        self.registrysize = unpacked_data[3]
        self.archivesize = unpacked_data[4]
        self.compressedregistry = bool(unpacked_data[5])
        self.reserved = unpacked_data[6:]


class FileRecord(dict):

    SIZEOFREC = 532

    def __init__(self, **kw):
        defaults = {
            'name': None,
            'path': None,
            'osize': 0,
            'csize': 0,
            'start': 0,
            'crc': 0,
            'compressed': False,
            'stored': False,
            'compressionlevel': 0
            }
        super(FileRecord, self).__init__(defaults, **kw)

    def __setattr__(self, key, value):
        self[key] = value

    def __getattr__(self, key):
        return self[key]

    def fromstream(self, stream):
        unpacked_data = _struct.unpack('<256B256BllllBBxx',
                                       stream.read(self.SIZEOFREC))
        self.name = _xgutils.delphishortstrtostr(unpacked_data[0:256])
        self.path = _xgutils.delphishortstrtostr(unpacked_data[256:512])
        self.osize = unpacked_data[512]
        self.csize = unpacked_data[513]
        self.start = unpacked_data[514]
        self.crc = unpacked_data[515] & 0xffffffff
        self.compressed = bool(unpacked_data[516] == 0)
        self.compressionlevel = unpacked_data[517]

    def __str__(self):
        return str(self.todict())


class ZlibArchive(object):
    __MAXBUFSIZE = 32768
    __TMP_PREFIX = 'tmpXGI'

    def __init__(self, stream=None, filename=None):
        self.arcrec = ArchiveRecord()
        self.arcregistry = []
        self.startofarcdata = -1
        self.endofarcdata = -1

        self.filename = filename
        self.stream = stream
        if stream is None:
            self.stream = open(filename, 'rb')

        self.__getarchiveindex()

    def __extractsegment(self, iscompressed=True, numbytes=None):
        # Extract a stored segment
        filename = None
        stream = []

        try:
            tmpfd, filename = _tempfile.mkstemp(prefix=self.__TMP_PREFIX)
            with _os.fdopen(tmpfd, "wb") as tmpfile:

                if (iscompressed):
                    # Extract a compressed segment
                    decomp = _zlib.decompressobj()
                    buf = self.stream.read(self.__MAXBUFSIZE)
                    stream = decomp.decompress(buf)

                    if len(stream) <= 0:
                        raise IOError()

                    tmpfile.write(stream)

                    # Read until we have uncompressed a complete segment
                    while len(decomp.unused_data) == 0:
                        block = self.stream.read(self.__MAXBUFSIZE)
                        if len(block) > 0:
                            try:
                                stream = decomp.decompress(block)
                                tmpfile.write(stream)
                            except:
                                break
                        else:
                            # EOF reached
                            break

                else:
                    # Extract an uncompressed segment
                    # Uncompressed segment needs numbytes specified
                    if numbytes is None:
                        raise IOError()

                    blksize = self.__MAXBUFSIZE
                    bytesleft = numbytes
                    while True:
                        if bytesleft < blksize:
                            blksize = bytesleft

                        block = self.stream.read(blksize)
                        tmpfile.write(block)
                        bytesleft = bytesleft - blksize

                        if bytesleft == 0:
                            break

        except (_zlib.error, IOError) as e:
            _os.unlink(filename)
            return None

        return filename

    def __getarchiveindex(self):

        try:
            # Advance to the archive record at the end and retrieve it
            filerecords = []
            curstreampos = self.stream.tell()

            self.stream.seek(-ArchiveRecord.SIZEOFREC, _os.SEEK_END)
            self.endofarcdata = self.stream.tell()
            self.arcrec.fromstream(self.stream)

            # Position ourselves at the beginning of the archive file index
            self.stream.seek(-ArchiveRecord.SIZEOFREC -
                             self.arcrec.registrysize, _os.SEEK_END)
            self.startofarcdata = self.stream.tell() - self.arcrec.archivesize

            # Compute the CRC32 of all the archive data including file index
            streamcrc = _xgutils.streamcrc32(
                    self.stream,
                    startpos=self.startofarcdata,
                    numbytes=(self.endofarcdata - self.startofarcdata))
            if streamcrc != self.arcrec.crc:
                raise Error("Archive CRC check failed - file corrupt")

            # Decompress the index into a temporary file
            idx_filename = self.__extractsegment(
                    iscompressed=self.arcrec.compressedregistry)
            if idx_filename is None:
                raise Error("Error extracting archive index")

            # Retrieve all the files in the index
            with open(idx_filename, "rb") as idx_file:
                for recordnum in range(0, self.arcrec.filecount):
                    curidxpos = self.stream.tell()

                    # Retrieve next file index record
                    filerec = FileRecord()
                    filerec.fromstream(idx_file)
                    filerecords.append(filerec)

                    self.stream.seek(curidxpos, 0)

            _os.unlink(idx_filename)
        finally:
            self.stream.seek(curstreampos, 0)

        self.arcregistry = filerecords

    def getarchivefile(self, filerec):
        # Do processing on the temporary file
        self.stream.seek(filerec.start + self.startofarcdata)
        tmpfilename = self.__extractsegment(iscompressed=filerec.compressed,
                                            numbytes=filerec.csize)
        if tmpfilename is None:
            raise Error("Error extracting archived file")
        tmpfile = open(tmpfilename, "rb")

        # Compute the CRC32 on the uncompressed file
        streamcrc = _xgutils.streamcrc32(tmpfile)
        if streamcrc != filerec.crc:
            raise Error("File CRC check failed - file corrupt")

        return tmpfile, tmpfilename

    def setblocksize(self, blksize):
        self.__MAXBUFSIZE = blksize


if __name__ == '__main__':
    pass



"""
Custom dialog for editing notes with word wrapping.
"""

from PySide6.QtWidgets import (
    QDialog, QVBoxLayout, QHBoxLayout, QPushButton,
    QLabel, QPlainTextEdit
)
from PySide6.QtCore import Qt


class NoteEditDialog(QDialog):
    """Custom dialog for editing notes with word wrapping."""

    def __init__(self, current_note: str = "", label_text: str = "Note:", parent=None):
        super().__init__(parent)
        self.setWindowTitle("Edit Note")
        self.setModal(True)
        self.setMinimumSize(500, 300)

        layout = QVBoxLayout(self)
        layout.setSpacing(12)

        # Label
        label = QLabel(label_text)
        label.setStyleSheet("color: #cdd6f4; font-weight: 600;")
        layout.addWidget(label)

        # Text edit with word wrap
        self.text_edit = QPlainTextEdit()
        self.text_edit.setPlainText(current_note)
        self.text_edit.setLineWrapMode(QPlainTextEdit.WidgetWidth)
        self.text_edit.setStyleSheet("""
            QPlainTextEdit {
                background-color: #1e1e2e;
                color: #cdd6f4;
                border: 2px solid #313244;
                border-radius: 6px;
                padding: 8px;
            }
            QPlainTextEdit:focus {
                border-color: #89b4fa;
            }
        """)
        layout.addWidget(self.text_edit, stretch=1)

        # Buttons
        button_layout = QHBoxLayout()
        button_layout.addStretch()

        self.btn_ok = QPushButton("OK")
        self.btn_ok.setStyleSheet("""
            QPushButton {
                background-color: #89b4fa;
                color: #1e1e2e;
                border: none;
                padding: 8px 24px;
                border-radius: 6px;
                font-weight: 600;
            }
            QPushButton:hover {
                background-color: #a0c8fc;
            }
        """)
        self.btn_ok.setCursor(Qt.PointingHandCursor)
        self.btn_ok.clicked.connect(self.accept)
        button_layout.addWidget(self.btn_ok)

        self.btn_cancel = QPushButton("Cancel")
        self.btn_cancel.setStyleSheet("""
            QPushButton {
                background-color: #45475a;
                color: #cdd6f4;
                border: none;
                padding: 8px 24px;
                border-radius: 6px;
            }
            QPushButton:hover {
                background-color: #585b70;
            }
        """)
        self.btn_cancel.setCursor(Qt.PointingHandCursor)
        self.btn_cancel.clicked.connect(self.reject)
        button_layout.addWidget(self.btn_cancel)

        layout.addLayout(button_layout)

        # Focus the text edit
        self.text_edit.setFocus()

    def get_text(self) -> str:
        """Get the edited text."""
        return self.text_edit.toPlainText()



#
#   xgstruct.py - classes to read XG file structures
#   Copyright (C) 2013,2014  Michael Petch <mpetch@gnubg.org>
#                                          <mpetch@capp-sysware.com>
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Lesser General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#
#   This code is based upon Delphi data structures provided by
#   Xavier Dufaure de Citres <contact@extremegammon.com> for purposes
#   of interacting with the ExtremeGammon XG file formats. Field
#   descriptions derived from xg_format.pas. The file formats are 
#   published at http://www.extremegammon.com/xgformat.aspx
#

from . import xgutils as _xgutils
import struct as _struct
import os as _os
import uuid as _uuid
import binascii as _binascii


class GameDataFormatHdrRecord(dict):
    SIZEOFREC = 8232

    def __init__(self, **kw):
        defaults = {
            'MagicNumber': 0,             # $484D4752, RM_MAGICNUMBER
            'HeaderVersion': 0,           # version
            'HeaderSize': 0,              # size of the header
            'ThumbnailOffset': 0,         # location of the thumbnail (jpg)
            'ThumbnailSize': 0,           # size in bye of the thumbnail
            'GameGUID': None,             # game id (GUID)
            'GameName': None,             # Unicode game name
            'SaveName': None,             # Unicode save name
            'LevelName': None,            # Unicode level name
            'Comments': None              # Unicode comments
            }                              
        super(GameDataFormatHdrRecord, self).__init__(defaults, **kw)

    def __setattr__(self, key, value):
        self[key] = value

    def __getattr__(self, key):
        return self[key]

    def fromstream(self, stream):
        try:
            unpacked_data = \
                    _struct.unpack('<4BiiQiLHHBB6s1024H1024H1024H1024H', 
                    stream.read(self.SIZEOFREC))
        except:
            return None

        self.MagicNumber = bytearray(unpacked_data[0:4][::-1]).decode('ascii')
        self.HeaderVersion = unpacked_data[4]
        if self.MagicNumber != 'HMGR' or self.HeaderVersion != 1:
            return None
            
        self.HeaderSize = unpacked_data[5]
        self.ThumbnailOffset = unpacked_data[6]
        self.ThumbnailSize = unpacked_data[7]
    
        # Convert Delphi 4 component GUID to the 6 components 
        # of a Python GUID.
        guidp1, guidp2, guidp3, guidp4, guidp5 = unpacked_data[8:13]
        guidp6 = int(_binascii.b2a_hex(unpacked_data[13]), 16)
        self.GameGUID = str(_uuid.UUID(fields=(guidp1, guidp2, guidp3,
                            guidp4, guidp5, guidp6)))

        self.GameName = _xgutils.utf16intarraytostr(unpacked_data[14:1038])
        self.SaveName = _xgutils.utf16intarraytostr(unpacked_data[1038:2062])
        self.LevelName = _xgutils.utf16intarraytostr(unpacked_data[2062:3086])
        self.Comments = _xgutils.utf16intarraytostr(unpacked_data[3086:4110])
        return self
        

class TimeSettingRecord(dict):

    SIZEOFREC = 32

    def __init__(self, **kw):
        defaults = {
            'ClockType': 0,                 # 0=None,0=Fischer,0=Bronstein
            'PerGame': False,               # time is for session reset after each game
            'Time1': 0,                     # initial time in sec
            'Time2': 0,                     # time added (fisher) or reverved (bronstrein) per move in sec
            'Penalty': 0,                   # point penalty when running our of time (in point)
            'TimeLeft1': 0,                 # current time left
            'TimeLeft2': 0,                 # current time left
            'PenaltyMoney': 0               # point penalty when running our of time (in point)
            }                              
        super(TimeSettingRecord, self).__init__(defaults, **kw)

    def __setattr__(self, key, value):
        self[key] = value

    def __getattr__(self, key):
        return self[key]

    def fromstream(self, stream):
        unpacked_data = _struct.unpack(
            '<lBxxxllllll',
            stream.read(self.SIZEOFREC))
        self.ClockType = unpacked_data[0]
        self.PerGame = bool(unpacked_data[1])
        self.Time1 = unpacked_data[2]
        self.Time2 = unpacked_data[3]
        self.Penalty = unpacked_data[4]
        self.TimeLeft1 = unpacked_data[5]
        self.TimeLeft2 = unpacked_data[6]
        self.PenaltyMoney = unpacked_data[7]
        return self


class EvalLevelRecord(dict):

    SIZEOFREC = 4

    def __init__(self, **kw):
        defaults = {
            'Level': 0,                     # Level used see PLAYERLEVEL table
            'isDouble': False               # The analyze assume double for the very next move
            }
        super(EvalLevelRecord, self).__init__(defaults, **kw)

    def __setattr__(self, key, value):
        self[key] = value

    def __getattr__(self, key):
       return self[key]

    def fromstream(self, stream):
        unpacked_data = _struct.unpack(
            '<hBb',
            stream.read(self.SIZEOFREC))
        self.Level = unpacked_data[0]
        self.isDouble = bool(unpacked_data[1])

        return self


class EngineStructBestMoveRecord(dict):

    SIZEOFREC = 2184

    def __init__(self, **kw):
        defaults = {
            'Pos': None,                    # Current position
            'Dice': None,                   # Dice
            'Level': 0,                     # analyze level requested
            'Score': None,                  # current score
            'Cube': 0,                      # cube value 1,2,4, etcc.
            'CubePos': 0,                   # 0: Center 1: Player owns cube -1 Opponent owns cube
            'Crawford': 0,                  # 1 = Crawford   0 = No Crawford
            'Jacoby': 0,                    # 1 = Jacoby   0 = No Jacoby
            'NMoves': 0,                    # number of move (max 32)
            'PosPlayed': None,              # position played
            'Moves': None,                  # move list as From1,dice1, from2,dice2 etc.. -1 show termination of list
            'EvalLevel': None,              # evaluation level of each move
            'Eval': None,                   # eval value of each move
            'Unused': 0,                    # if 1 does not count as a decision
            'met': 0,                       # unused
            'Choice0': 0,                   # 1-ply choice (index to PosPlayed)
            'Choice3': 0                    # 3-ply choice (index to PosPlayed)
            }
        super(EngineStructBestMoveRecord, self).__init__(defaults, **kw)

    def __setattr__(self, key, value):
        self[key] = value

    def __getattr__(self, key):
        return self[key]

    def fromstream(self, stream):
        unpacked_data = _struct.unpack(
                '<26bxx2ll2llllll',
                stream.read(68))
        self.Pos = unpacked_data[0:26]
        self.Dice = unpacked_data[26:28]
        self.Level = unpacked_data[28]
        self.Score = unpacked_data[29:31]
        self.Cube = unpacked_data[31]
        self.Cubepos = unpacked_data[32]
        self.Crawford = unpacked_data[33]
        self.Jacoby = unpacked_data[34]
        self.NMoves = unpacked_data[35]

        self.PosPlayed = ()
        for row in range(32):
            unpacked_data = _struct.unpack('<26b', stream.read(26))
            self.PosPlayed += (unpacked_data[0:26],)

        self.Moves = ()
        for row in range(32):
            self.Moves += (_struct.unpack('<8b', stream.read(8))[0:8],)

        self.EvalLevel = ()
        for row in range(32):
            self.EvalLevel += (EvalLevelRecord().fromstream(stream),)

        self.Eval = ()
        for row in range(32):
            unpacked_data = _struct.unpack('<7f', stream.read(28))
            self.Eval += (unpacked_data,)

        unpacked_data = _struct.unpack('<bbbb', stream.read(4))
        self.Unused = unpacked_data[0]
        self.met = unpacked_data[1]
        self.Choice0 = unpacked_data[2]
        self.Choice3 = unpacked_data[3]

        return self


class EngineStructDoubleAction(dict):

    SIZEOFREC = 132

    def __init__(self, **kw):
        defaults = {
            'Pos': None,                    # Current position
            'Level': 0,                     # analyze level performed
            'Score': None,                  # current score
            'Cube': 0,                      # cube value 1,2,4, etcc.
            'CubePos': 0,                   # 0: Center 1: Player owns cube -1 Opponent owns cube
            'Jacoby': 0,                    # 1 = Jacoby   0 = No Jacoby
            'Crawford': 0,                  # 1 = Crawford   0 = No Crawford
            'met': 0,                       # unused
            'FlagDouble': 0,                # 0: Dont double 1: Double
            'isBeaver': 0,                  # is it a beaver if doubled
            'Eval': None,                   # eval value for No double
            'equB': 0.0,                    # equity No Double
            'equDouble': 0.0,               # equity Double/take
            'equDrop': 0.0,                 # equity double/drop (-1)
            'LevelRequest': 0,              # analyze level requested
            'DoubleChoice3': 0,             # 3-ply choice as double+take*2
            'EvalDouble': None              # eval value for Double/Take
            }
        super(EngineStructDoubleAction, self).__init__(defaults, **kw)

    def __setattr__(self, key, value):
        self[key] = value

    def __getattr__(self, key):
       return self[key]

    def fromstream(self, stream):
        unpacked_data = _struct.unpack(
                '<26bxxl2llllhhhh7ffffhh7f',
                stream.read(132))
        self.Pos = unpacked_data[0:26]
        self.Level = unpacked_data[26]
        self.Score = unpacked_data[27:29]
        self.Cube = unpacked_data[29]
        self.CubePos = unpacked_data[30]
        self.Jacoby = unpacked_data[31]
        self.Crawford = unpacked_data[32]
        self.met = unpacked_data[33]
        self.FlagDouble = unpacked_data[34]
        self.isBeaver = unpacked_data[35]
        self.Eval = unpacked_data[36:43]
        self.equB = unpacked_data[43]
        self.equDouble = unpacked_data[44]
        self.equDrop = unpacked_data[45]
        self.LevelRequest = unpacked_data[46]
        self.DoubleChoice3 = unpacked_data[47]
        self.EvalDouble = unpacked_data[48:55]

        return self

class HeaderMatchEntry(dict):

    SIZEOFREC = 2560

    def __init__(self, version=0, **kw):
        defaults = {
            'Name': 'MatchInfo',
            'EntryType': GameFileRecord.ENTRYTYPE_HEADERMATCH,
            'SPlayer1': None,              # player name in ANSI string for XG1 compatbility see "Player1" and "Player2" below for unicode
            'SPlayer2': None,
            'MatchLength': 0,              # Match length, 99999 for unlimited
            'Variation': 0,                # 0:backgammon, 1: Nack, 2: Hyper, 3: Longgammon
            'Crawford': False,             # Crawford in use
            'Jacoby': False,               # Jacoby in use
            'Beaver': False,               # Beaver in use
            'AutoDouble': False,           # Automatic double in use
            'Elo1': 0.0,                   # player1 elo
            'Elo2': 0.0,                   # player2 experience
            'Exp1': 0,                     # player1 elo
            'Exp2': 0,                     # player2 experience
            'Date': 0,                     # game date
            'SEvent': None,                # event name, in ANSI string for XG1 compatbility see "event" below for unicode
            'GameId': 0,                   # game ID, if player are swap make gameid:=-GameID
            'CompLevel1': -1,              # Player level: see table at the end (PLAYERLEVEL TABLE)
            'CompLevel2': -1,
            'CountForElo': False,          # outcome of the session will affect elo
            'AddtoProfile1': False,        # outcome of the session will affect player 1 profile
            'AddtoProfile2': False,        # outcome of the session will affect player 2 profile
            'SLocation': None,             # location name, in ANSI string for XG1 compatbility see "location" below for unicode
            'GameMode': 0,                 # game mode : see table at the end (GAMEMODE TABLE)
            'Imported': False,             # game was imported from an site (MAT, CBG etc..)
            'SRound': None,                # round name, in ANSI string for XG1 compatbility see "round" below for unicode
            'Invert': 0,                   # If the board is swap then invert=-invert and MatchID=-MatchID
            'Version': version,            # file version, currently SaveFileVersion
            'Magic': 0x494C4D44,           # must be MagicNumber = $494C4D44;
            'MoneyInitG': 0,               # initial game played from the profile against that opp in money
            'MoneyInitScore': [0, 0],      # initial score from the profile against that opp in money
            'Entered': False,              # entered in profile
            'Counted': False,              # already accounted in the profile elo
            'UnratedImp': False,           # game was unrated on the site it was imported from
            'CommentHeaderMatch': -1,      # index of the match comment header in temp.xgc
            'CommentFooterMatch': -1,      # index of the match comment footer in temp.xgc
            'isMoneyMatch': False,         # was player for real money
            'WinMoney': 0.0,               # amount of money for the winner
            'LoseMoney': 0.0,              # amount of money for the looser
            'Currency': 0,                 # currency code from Currency.ini
            'FeeMoney': 0.0,               # amount of rake
            'TableStake': 0,               # max amount that can be lost -- NOT IMPLEMENTED
            'SiteId': -1,                  # site id from siteinfo.ini
            'CubeLimit': 0,                # v8: maximum cube value
            'AutoDoubleMax': 0,            # v8: maximum c# of time the autodouble can be used
            'Transcribed': False,          # v24: game was transcribed
            'Event': None,                 # v24: Event name (unicode)
            'Player1': None,               # v24: Player1 name (unicode)
            'Player2': None,               # v24: Player2 name (unicode)
            'Location': None,              # v24: Location (unicode)
            'Round': None,                 # v24: Round (unicode)
            'TimeSetting': None,           # v25: Time setting for the game
            'TotTimeDelayMove': 0,         # v26: # of checker play marked for delayed RO
            'TotTimeDelayCube': 0,         # v26: # of checker play marked for delayed RO done
            'TotTimeDelayMoveDone': 0,     # v26: # of checker Cube action marked for delayed RO
            'TotTimeDelayCubeDone': 0,     # v26: # of checker Cube action marked for delayed RO Done
            'Transcriber': None            # v30: Name of the Transcriber (unicode)
            }
        super(HeaderMatchEntry, self).__init__(defaults, **kw)

    def __setattr__(self, key, value):
        self[key] = value

    def __getattr__(self, key):
       return self[key]

    def fromstream(self, stream):

        unpacked_data = _struct.unpack(
            '<9x41B41BxllBBBBddlld129BxxxlllBBB129BlB129BxxllLl2lBBB'
            'xllBxxxfflfll', stream.read(612))
        self.SPlayer1 = _xgutils.delphishortstrtostr(unpacked_data[0:41])
        self.SPlayer2 = _xgutils.delphishortstrtostr(unpacked_data[41:82])
        self.MatchLength = unpacked_data[82]
        self.Variation = unpacked_data[83]
        self.Crawford = bool(unpacked_data[84])
        self.Jacoby = bool(unpacked_data[85])
        self.Beaver = bool(unpacked_data[86])
        self.AutoDouble = bool(unpacked_data[87])
        self.Elo1 = unpacked_data[88]
        self.Elo2 = unpacked_data[89]
        self.Exp1 = unpacked_data[90]
        self.Exp2 = unpacked_data[91]
        self.Date = str(_xgutils.delphidatetimeconv(unpacked_data[92]))
        self.SEvent = _xgutils.delphishortstrtostr(unpacked_data[93:222])
        self.GameId = unpacked_data[222]
        self.CompLevel1 = unpacked_data[223]
        self.CompLevel2 = unpacked_data[224]
        self.CountForElo = bool(unpacked_data[225])
        self.AddtoProfile1 = bool(unpacked_data[226])
        self.AddtoProfile2 = bool(unpacked_data[227])
        self.SLocation = _xgutils.delphishortstrtostr(unpacked_data[228:357])
        self.GameMode = unpacked_data[357]
        self.Imported = bool(unpacked_data[358])
        self.SRound = _xgutils.delphishortstrtostr(unpacked_data[359:487])
        self.Invert = unpacked_data[488]
        self.Version = unpacked_data[489]
        self.Magic = unpacked_data[490]
        self.MoneyInitG = unpacked_data[491]
        self.MoneyInitScore = unpacked_data[492:494]
        self.Entered = bool(unpacked_data[494])
        self.Counted = bool(unpacked_data[495])
        self.UnratedImp = bool(unpacked_data[496])
        self.CommentHeaderMatch = unpacked_data[497]
        self.CommentFooterMatch = unpacked_data[498]
        self.isMoneyMatch = bool(unpacked_data[499])
        self.WinMoney = unpacked_data[500]
        self.LoseMoney = unpacked_data[501]
        self.Currency = unpacked_data[502]
        self.FeeMoney = unpacked_data[503]
        self.TableStake = unpacked_data[504]
        self.SiteId = unpacked_data[505]
        if self.Version >= 8:
            unpacked_data = _struct.unpack('<ll', stream.read(8))
            self.CubeLimit = unpacked_data[0]
            self.AutoDoubleMax = unpacked_data[1]
        if self.Version >= 24:
            unpacked_data = _struct.unpack('<Bx129H129H129H129H129H',
                                           stream.read(1292))
            self.Transcribed = bool(unpacked_data[0])
            self.Event = _xgutils.utf16intarraytostr(unpacked_data[1:130])
            self.Player1 = _xgutils.utf16intarraytostr(unpacked_data[130:259])
            self.Player2 = _xgutils.utf16intarraytostr(unpacked_data[259:388])
            self.Location = _xgutils.utf16intarraytostr(unpacked_data[388:517])
            self.Round = _xgutils.utf16intarraytostr(unpacked_data[517:646])
        if self.Version >= 25:
            self.TimeSetting = TimeSettingRecord().fromstream(stream)
        if self.Version >= 26:
            unpacked_data = _struct.unpack('<llll', stream.read(16))
            self.TotTimeDelayMove = unpacked_data[0]
            self.TotTimeDelayCube = unpacked_data[1]
            self.TotTimeDelayMoveDone = unpacked_data[2]
            self.TotTimeDelayCubeDone = unpacked_data[3]
        if self.Version >= 30:
            unpacked_data = _struct.unpack('<129H', stream.read(258))
            self.Transcriber = _xgutils.utf16intarraytostr(
                unpacked_data[0:129])

        return self


class FooterGameEntry(dict):

    SIZEOFREC = 2560

    def __init__(self, **kw):
        defaults = {
            'Name': 'GameFooter',
            'EntryType': GameFileRecord.ENTRYTYPE_FOOTERGAME,
            'Score1g': 0,                   # Final score
            'Score2g': 0,                   # Final score
            'CrawfordApplyg': False,        # will crawford apply next game
            'Winner': 0,                    # who win +1=player1, -1 player 2
            'PointsWon': 0,                 # point scored
            'Termination': 0,               # 0=Drop 1=single 2=gammon 3=Backgamon 
                                            # (0,1,2)+100=Resign  (0,1,2)+1000 settle
            'ErrResign': 0.0,               # error made by resigning (-1000 if not analyze)
            'ErrTakeResign': 0.0,           # error made by accepting the resign (-1000 if not analyze)
            'Eval': None,                   # evaluation of the final position
            'EvalLevel': 0
            }
        super(FooterGameEntry, self).__init__(defaults, **kw)

    def __setattr__(self, key, value):
        self[key] = value

    def __getattr__(self, key):
       return self[key]

    def fromstream(self, stream):
        unpacked_data = _struct.unpack('<9xxxxllBxxxlllxxxxdd7dl',
                stream.read(116))
        self.Score1g = unpacked_data[0]
        self.Score2g = unpacked_data[1]
        self.CrawfordApplyg = bool(unpacked_data[2])
        self.Winner = unpacked_data[3]
        self.PointsWon = unpacked_data[4]
        self.Termination = unpacked_data[5]
        self.ErrResign = unpacked_data[6]
        self.ErrTakeResign = unpacked_data[7]
        self.Eval = unpacked_data[8:15]
        self.EvalLevel = unpacked_data[15]
        return self


class MissingEntry(dict):

    SIZEOFREC = 2560

    def __init__(self, **kw):
        defaults = {
            'Name': 'Missing',
            'EntryType': GameFileRecord.ENTRYTYPE_MISSING,
            'MissingErrLuck': 0.0,
            'MissingWinner': 0,
            'MissingPoints': 0
            }
        super(MissingEntry, self).__init__(defaults, **kw)

    def __setattr__(self, key, value):
        self[key] = value

    def __getattr__(self, key):
       return self[key]

    def fromstream(self, stream):
        unpacked_data = _struct.unpack('<9xxxxxxxxdll', stream.read(32))
        self.MissingErrLuck = unpacked_data[0]
        self.MissingWinner = unpacked_data[1]
        self.MissingPoints = unpacked_data[2]
        return self


class FooterMatchEntry(dict):

    SIZEOFREC = 2560

    def __init__(self, **kw):
        defaults = {
            'Name': 'MatchFooter',
            'EntryType': GameFileRecord.ENTRYTYPE_FOOTERMATCH,
            'Score1m': 0,                   # Final score of the match
            'Score2m': 0,                   # Final score of the match
            'WinnerM': 0,                   # who win +1=player1, -1 player 2
            'Elo1m': 0.0,                   # resulting elo, player1
            'Elo2m': 0.0,                   # resulting elo, player2
            'Exp1m': 0,                     # resulting exp, player1
            'Exp2m': 0,                     # resulting exp, player2
            'Datem': 0.0                    # Date time of the match end
            }
        super(FooterMatchEntry, self).__init__(defaults, **kw)

    def __setattr__(self, key, value):
        self[key] = value

    def __getattr__(self, key):
       return self[key]

    def fromstream(self, stream):
        unpacked_data = _struct.unpack('<9xxxxlllddlld', stream.read(56))
        self.Score1m = unpacked_data[0]
        self.Score2m = unpacked_data[1]
        self.WinnerM = unpacked_data[2]
        self.Elo1m = unpacked_data[3]
        self.Elo2m = unpacked_data[4]
        self.Exp1m = unpacked_data[5]
        self.Exp2m = unpacked_data[6]
        self.Datem = str(_xgutils.delphidatetimeconv(unpacked_data[7]))

        return self


class HeaderGameEntry(dict):

    SIZEOFREC = 2560

    def __init__(self, **kw):
        defaults = {
            'Name': 'GameHeader',
            'EntryType': GameFileRecord.ENTRYTYPE_HEADERGAME,
            'Score1': 0,                    # initial score player1
            'Score2': 0,                    # initial score player1
            'CrawfordApply': False,         # iDoes Crawford apply on that game
            'PosInit': (0,) * 26,           # initial position
            'GameNumber': 0,                # Game number (start at 1)
            'InProgress': False,            # Game is still in progress
            'CommentHeaderGame': -1,        # index of the game comment header in temp.xgc
            'CommentFooterGame': -1,        # index of the game comment footer in temp.xgc
            'NumberOfAutoDoubles': 0        # v26: Number of Autodouble that happen in that game
                                            # note that in the rest of the game the cube still start at 1.
                                            # For display purpose or point calculation add the 2^NumberOfAutoDouble
            }
        super(HeaderGameEntry, self).__init__(defaults, **kw)

    def __setattr__(self, key, value):
        self[key] = value

    def __getattr__(self, key):
       return self[key]

    def fromstream(self, stream):
        unpacked_data = _struct.unpack('<9xxxxllB26bxlBxxxlll',
                stream.read(68))
        self.Score1 = unpacked_data[0]
        self.Score2 = unpacked_data[1]
        self.CrawfordApply = bool(unpacked_data[2])
        self.PosInit = unpacked_data[3:29]
        self.GameNumber = unpacked_data[29]
        self.InProgress = bool(unpacked_data[30])
        self.CommentHeaderGame = unpacked_data[31]
        self.CommentFooterGame = unpacked_data[32]
        if self.Version >= 26:
            self.NumberOfAutoDoubles = unpacked_data[33]

        return self


class CubeEntry(dict):

    SIZEOFREC = 2560

    def __init__(self, **kw):
        defaults = {
            'Name': 'Cube',
            'EntryType': GameFileRecord.ENTRYTYPE_CUBE,
            'ActiveP': 0,                   # Active player (1 or 2)
            'Double': 0,                    # player double (0= no, 1=yes)
            'Take': 0,                      # opp take (0= no, 1=yes, 2=beaver )
            'BeaverR': 0,                   # player accept beaver (0= no, 1=yes, 2=raccoon)
            'RaccoonR': 0,                  # player accept raccoon (0= no, 1=yes)
            'CubeB': 0,                     # Cube value 0=center, +1=2 own, +2=4 own ... -1=2 opp, -2=4 opp
            'Position': None,               # initial position
            'Doubled': None,                # Analyze result
            'ErrCube': 0.0,                 # error made on doubling (-1000 if not analyze)
            'DiceRolled': None,             # dice rolled
            'ErrTake': 0.0,                 # error made on taking (-1000 if not analyze)
            'RolloutIndexD': 0,             # index of the Rollout in temp.xgr
            'CompChoiceD': 0,               # 3-ply choice as Double+2*take
            'AnalyzeC': 0,                  # Level of the analyze
            'ErrBeaver': 0.0,               # error made on beavering (-1000 if not analyze)
            'ErrRaccoon': 0.0,              # error made on racconning (-1000 if not analyze)
            'AnalyzeCR': 0,                 # requested Level of the analyze (sometime a XGR+ request will stop at 4-ply when obivous)
            'isValid': 0,                   # invalid decision 0=Ok, 1=error, 2=invalid
            'TutorCube': 0,                 # player initial double in tutor mode (0= no, 1=yes)
            'TutorTake': 0,                 # player initial take in tutor mode (0= no, 1=yes)
            'ErrTutorCube': 0.0,            # error initialy made on doubling (-1000 if not analyze)
            'ErrTutorTake': 0.0,            # error initialy made on taking (-1000 if not analyze)
            'FlaggedDouble': False,         # cube has been flagged
            'CommentCube': -1,              # index of the cube comment in temp.xgc
            'EditedCube': False,            # v24: Position was edited at this point
            'TimeDelayCube': False,         # v26: position is marked for later RO
            'TimeDelayCubeDone': False,     # v26: position later RO has been done
            'NumberOfAutoDoubleCube': 0,    # v27: Number of Autodouble that happen in that game
            'TimeBot': 0,                   # v28: time left for both players
            'TimeTop': 0
            }
        super(CubeEntry, self).__init__(defaults, **kw)

    def __setattr__(self, key, value):
        self[key] = value

    def __getattr__(self, key):
       return self[key]

    def fromstream(self, stream):
        unpacked_data = _struct.unpack('<9xxxxllllll26bxx',
                                       stream.read(64))
        self.ActiveP = unpacked_data[0]
        self.Double = unpacked_data[1]
        self.Take = unpacked_data[2]
        self.BeaverR = unpacked_data[3]
        self.RaccoonR = unpacked_data[4]
        self.CubeB = unpacked_data[5]
        self.Position = unpacked_data[6:32]
        self.Doubled = EngineStructDoubleAction().fromstream(stream)
        unpacked_data = _struct.unpack('<xxxxd3Bxxxxxdlllxxxx' \
                                       'ddllbbxxxxxxddBxxxlBBBxlll',
                                       stream.read(116))
        self.ErrCube = unpacked_data[0]
        self.DiceRolled = _xgutils.delphishortstrtostr(unpacked_data[1:4])
        self.ErrTake = unpacked_data[4]
        self.RolloutIndexD = unpacked_data[5]
        self.CompChoiceD = unpacked_data[6]
        self.AnalyzeC = unpacked_data[7]
        self.ErrBeaver = unpacked_data[8]
        self.ErrRaccoon = unpacked_data[9]
        self.AnalyzeCR = unpacked_data[10]
        self.isValid = unpacked_data[11]
        self.TutorCube = unpacked_data[12]
        self.TutorTake = unpacked_data[13]
        self.ErrTutorCube = unpacked_data[14]
        self.ErrTutorTake = unpacked_data[15]
        self.FlaggedDouble = bool(unpacked_data[16])
        self.CommentCube = unpacked_data[17]
        if self.Version >= 24:
            self.EditedCube = bool(unpacked_data[18])
        if self.Version >= 26:
            self.TimeDelayCube = bool(unpacked_data[19])
            self.TimeDelayCubeDone = bool(unpacked_data[20])
        if self.Version >= 27:
            self.NumberOfAutoDoubleCube = unpacked_data[21]
        if self.Version >= 28:
            self.TimeBot = unpacked_data[22]
            self.TimeTop = unpacked_data[23]
        return self


class MoveEntry(dict):

    SIZEOFREC = 2560

    def __init__(self, **kw):
        defaults = {
            'Name:': 'Move',
            'EntryType': GameFileRecord.ENTRYTYPE_MOVE,
            'PositionI': None,              # Initial position
            'PositionEnd': None,            # Final Position
            'ActiveP': 0,                   # active player (1,2)
            'Moves': None,                  # list of move as From1,dice1, from2,dice2 etc.. -1 show termination of list
            'Dice': None,                   # dice rolled
            'CubeA': 0,                     # Cube value 0=center, +1=2 own, +2=4 own ... -1=2 opp, -2=4 opp
            'ErrorM': 0,                    # Not used anymore (not sure)
            'NMoveEval': 0,                 # Number of candidate (max 32)
            'DataMoves': None,              # analyze
            'Played': False,                # move was played
            'ErrMove': 0.0,                 # error made (-1000 if not analyze)
            'ErrLuck': 0.0,                 # luck of the roll
            'CompChoice': 0,                # computer choice (index to DataMoves.moveplayed)
            'InitEq': 0.0,                  # Equity before the roll (for luck purposes)
            'RolloutIndexM': None,          # index of the Rollout in temp.xgr
            'AnalyzeM': 0,                  # level of analyze of the move
            'AnalyzeL': 0,                  # level of analyze for the luck
            'InvalidM': 0,                  # invalid decision 0=Ok, 1=error, 2=invalid
            'PositionTutor': None,          # Position resulting of the initial move
            'Tutor': 0,                     # index of the move played dataMoves.moveplayed
            'ErrTutorMove': 0.0,            # error initialy made (-1000 if not analyze)
            'Flagged': False,               # move has been flagged
            'CommentMove': -1,              # index of the move comment in temp.xgc
            'EditedMove': False,            # v24: Position was edited at this point
            'TimeDelayMove': 0,             # v26: Bit list: position is marked for later RO
            'TimeDelayMoveDone': 0,         # v26: Bit list: position later RO has been done
            'NumberOfAutoDoubleMove': 0     # v27: Number of Autodouble that happen in that game
            }
        super(MoveEntry, self).__init__(defaults, **kw)

    def __setattr__(self, key, value):
        self[key] = value

    def __getattr__(self, key):
       return self[key]

    def fromstream(self, stream):
        unpacked_data = _struct.unpack('<9x26b26bxxxl8l2lldl',
                                       stream.read(124))
        self.PositionI = unpacked_data[0:26]
        self.PositionEnd = unpacked_data[26:52]
        self.ActiveP = unpacked_data[52]
        self.Moves = unpacked_data[53:61]
        self.Dice = unpacked_data[61:63]
        self.CubeA = unpacked_data[63]
        self.ErrorM = unpacked_data[64] # Not used
        self.NMoveEval = unpacked_data[65]
        self.DataMoves = EngineStructBestMoveRecord().fromstream(stream)

        unpacked_data = _struct.unpack('<Bxxxddlxxxxd32llll26bbxdBxxxl',
                                       stream.read(220))
        self.Played = bool(unpacked_data[0])
        self.ErrMove = unpacked_data[1]
        self.ErrLuck = unpacked_data[2]
        self.CompChoice = unpacked_data[3]
        self.InitEq = unpacked_data[4]
        self.RolloutIndexM = unpacked_data[5:37]
        self.AnalyzeM = unpacked_data[37]
        self.AnalyzeL = unpacked_data[38]
        self.InvalidM = unpacked_data[39]
        self.PositionTutor = unpacked_data[40:66]
        self.Tutor = unpacked_data[66]
        self.ErrTutorMove = unpacked_data[67]
        self.Flagged = bool(unpacked_data[68])
        self.CommentMove = unpacked_data[69]
        if self.Version >= 24:
            unpacked_data = _struct.unpack('<B', stream.read(1))
            self.EditedMove = bool(unpacked_data[0])
        if self.Version >= 26:
            unpacked_data = _struct.unpack('<xxxLL', stream.read(11))
            self.TimeDelayMove = unpacked_data[0]
            self.TimeDelayMoveDone = unpacked_data[1]
        if self.Version >= 27:
            unpacked_data = _struct.unpack('<l', stream.read(4))
            self.NumberOfAutoDoubleMove = unpacked_data[0]

        return self


class UnimplementedEntry(dict):

    """ Class for record types we have yet to implement
    """

    SIZEOFREC = 2560

    def __init__(self, **kw):
        defaults = {
            'Name': 'Unimplemented'
            }
        super(UnimplementedEntry, self).__init__(defaults, **kw)

    def __setattr__(self, key, value):
        self[key] = value

    def __getattr__(self, key):
       return self[key]

    def fromstream(self, stream):
        return self


class GameFileRecord(dict):

    __SIZEOFSRHDR = 9
    __REC_CLASSES = [HeaderMatchEntry, HeaderGameEntry,
                     CubeEntry, MoveEntry,
                     FooterGameEntry, FooterMatchEntry,
                     UnimplementedEntry, MissingEntry]

    ENTRYTYPE_HEADERMATCH, ENTRYTYPE_HEADERGAME, ENTRYTYPE_CUBE, \
            ENTRYTYPE_MOVE, ENTRYTYPE_FOOTERGAME, ENTRYTYPE_FOOTERMATCH, \
            ENTRYTYPE_MISSING, ENTRYTYPE_UNIMPLEMENTED = range(8)

    def __init__(self, version=-1, **kw):
        """ Create a game file record based upon the given file version
        number. The file version is first found in a HeaderMatchEntry
        object. The version needs to be propogated to all other game
        file objects within the same archive.
        """
        defaults = {
            'Name': 'GameFileRecord',
            'EntryType': -1,
            'Record': None,
            'Version': version
            }
        super(GameFileRecord, self).__init__(defaults, **kw)

    def __setattr__(self, key, value):
        self[key] = value

    def __getattr__(self, key):
       return self[key]

    def fromstream(self, stream):
        # Read the header. First 8 bytes are unused. 9th byte is record type
        # The record type determines what object to create and load.
        # If we catch a struct.error we have hit the EOF.
        startpos = stream.tell()
        try:
            unpacked_data = _struct.unpack('<8xB',
                                           stream.read(self.__SIZEOFSRHDR))
        except _struct.error:
            return None
        self.EntryType = unpacked_data[0]

        # Back up to the beginning of the record after getting the record
        # type and feed the entire stream back into the corresponding
        # record object.
        stream.seek(-self.__SIZEOFSRHDR, _os.SEEK_CUR)

        # Using the appropriate class, read the data stream
        self.Record = self.__REC_CLASSES[self.EntryType]()
        self.Record.Version = self.Version
        self.Record.fromstream(stream)
        realrecsize = stream.tell() - startpos

        # Each record is actually 2560 bytes long. We need to advance past
        # the unused filler data to be at the start of the next record
        stream.seek(self.Record.SIZEOFREC - realrecsize, _os.SEEK_CUR)

        return self.Record


class RolloutContextEntry(dict):

    SIZEOFREC = 2184

    def __init__(self, **kw):
        defaults = {
            'Name': 'Rollout',
            'EntryType': RolloutFileRecord.ROLLOUTCONTEXT,
            'Truncated': False,             # is truncated
            'ErrorLimited': False,          # stop when CI under "ErrorLimit"
            'Truncate': 0,                  # truncation level
            'MinRoll': 0,                   # minimum games to roll
            'ErrorLimit': 0.0,              # CI to stop the RO
            'MaxRoll': 0,                   # maximum games to roll
            'Level1': 0,                    # checker play Level used before "LevelCut"
            'Level2': 0,                    # checker play Level used on and after "LevelCut"
            'LevelCut': 0,                  # Cutoff for level1 and level2
            'Variance': False,              # use variance reduction
            'Cubeless': False,              # is a cubeless ro
            'Time': False,                  # is time limited
            'Level1C': 0,                   # cube Level used before "LevelCut"
            'Level2C': 0,                   # cube Level used on and after "LevelCut"
            'TimeLimit': 0,                 # limit in time (min)
            'TruncateBO': 0,                # what do do when reaching BO db: 0=nothing; 1=?
            'RandomSeed': 0,                # caculated seed=RandomSeedI+hashpos
            'RandomSeedI': 0,               # used entered seed
            'RollBoth': False,              # roll both line (ND and D/T)
            'SearchInterval': 0.0,          # Search interval used (1=normal, 1.5=large, 2=huge, 4=gigantic)
            'met': 0,                       # unused
            'FirstRoll': False,             # is it a first roll rollout
            'DoDouble': False,              # roll both line (ND and D/T) in multiple rollout
            'Extent': False,                # if the ro is extended
            'Rolled': 0,                    # game rolled
            'DoubleFirst': False,           # a double happens immediatly.
            'Sum1': None,                   # sum of equities for all 36 1st roll
            'SumSquare1': None,             # sum of square equities for all 36 1st roll
            'Sum2': None,                   # D/T sum of equities for all 36 1st roll
            'SumSquare2': None,             # D/T sum of square equities for all 36 1st roll
            'Stdev1': None,                 # Standard deviation for all 36 1st roll
            'Stdev2': None,                 # D/T Stand deviation for all 36 1st roll
            'RolledD': None,                # number of game rolled for all 36 1st roll
            'Error1': 0.0,                  # 95% CI
            'Error2': 0.0,                  # D/T 95% CI
            'Result1': None,                # evaluation of the position
            'Result2': None,                # D/T evaluation of the position
            'Mwc1': 0.0,                    # ND  mwc equivalent of result1[1,6]
            'Mwc2': 0.0,                    # D/T mwc equivalent of result2[1,6]
            'PrevLevel': 0,                 # store the previous analyze level (for deleting RO)
            'PrevEval': None,               # store the previous analyze result (for deleting RO)
            'PrevND': 0,                  # store the previous analyze equities (for deleting RO)
            'PrevD': 0,                   
            'Duration': 0,                # duration in seconds
            'LevelTrunc': 0,                # level used at truncation
            'Rolled2': 0,                   # D/T number of game rolled
            'MultipleMin': 0,               # Multiple RO minimum # of game
            'MultipleStopAll': False,       # Multiple RO stop all if one move reach MultipleStopAllValue
            'MultipleStopOne': False,       # Multiple RO stop one move is reach under MultipleStopOneValue
            'MultipleStopAllValue': 0.0,    # value to stop all RO (for instance 99.9%)
            'MultipleStopOneValue': 0.0,    # value to stop one move(for instance 0.01%)
            'AsTake': False,                # when running ND and D/T if AsTake is true, checker decision are made using the cube position in the D/T line
            'Rotation': 0,                  # 0=36 dice, 1=21 dice (XG1), 2=30 dice (for 1st pos)
            'UserInterrupted': False,       # RO was interrupted by user
            'VerMaj': 0,                    # Major version use for the RO, currently (2.20): 2
            'VerMin': 0                     # Minor version use for the RO, currently (2.10): 10 (no change in RO or engine between 2.10 and 2.20)
            }
        super(RolloutContextEntry, self).__init__(defaults, **kw)

    def __setattr__(self, key, value):
        self[key] = value

    def __getattr__(self, key):
       return self[key]

    def fromstream(self, stream):
        unpacked_data = _struct.unpack('<BBxxllxxxxdllllBBBxllLlllBxxx' \
                                       'flBBBxlBxxxxxxx37d37d37d37d37d37d37l' \
                                       'ff7f7fffl7fllllllBBxxffBxxxlBxHH',
                                       stream.read(2174))

        self.Truncated = bool(unpacked_data[0])
        self.ErrorLimited = bool(unpacked_data[1])
        self.Truncate = unpacked_data[2]
        self.MinRoll = unpacked_data[3]
        self.ErrorLimit = unpacked_data[4]
        self.MaxRoll = unpacked_data[5]
        self.Level1 = unpacked_data[6]
        self.Level2 = unpacked_data[7]
        self.LevelCut = unpacked_data[8]
        self.Variance = bool(unpacked_data[9])
        self.Cubeless = bool(unpacked_data[10])
        self.Time = bool(unpacked_data[11])
        self.Level1C = unpacked_data[12]
        self.Level2C = unpacked_data[13]
        self.TimeLimit = unpacked_data[14]
        self.TruncateBO = unpacked_data[15]
        self.RandomSeed = unpacked_data[16]
        self.RandomSeedI = unpacked_data[17]
        self.RollBoth = bool(unpacked_data[18])
        self.SearchInterval = unpacked_data[19]
        self.met = unpacked_data[20]
        self.FirstRoll = bool(unpacked_data[21])
        self.DoDouble = bool(unpacked_data[22])
        self.Extent = bool(unpacked_data[23])
        self.Rolled = unpacked_data[24]
        self.DoubleFirst = bool(unpacked_data[25])
        self.Sum1 = unpacked_data[26:63]
        self.SumSquare1 = unpacked_data[63:100]
        self.Sum2 = unpacked_data[100:137]
        self.SumSquare2 = unpacked_data[137:174]
        self.Stdev1 = unpacked_data[174:211]
        self.Stdev2 = unpacked_data[211:248]
        self.RolledD = unpacked_data[248:285]
        self.Error1 = unpacked_data[285]
        self.Error2 = unpacked_data[286]
        self.Result1 = unpacked_data[287:294]
        self.Result2 = unpacked_data[294:301]
        self.Mwc1 = unpacked_data[301]
        self.Mwc2 = unpacked_data[302]
        self.PrevLevel = unpacked_data[303]
        self.PrevEval = unpacked_data[304:311]
        self.PrevND = unpacked_data[311]
        self.PrevD = unpacked_data[312]
        self.Duration = unpacked_data[313]
        self.LevelTrunc = unpacked_data[314]
        self.Rolled2 = unpacked_data[315]
        self.MultipleMin = unpacked_data[316]
        self.MultipleStopAll = bool(unpacked_data[317])
        self.MultipleStopOne = bool(unpacked_data[318])
        self.MultipleStopAllValue = unpacked_data[319]
        self.MultipleStopOneValue = unpacked_data[320]
        self.AsTake = bool(unpacked_data[321])
        self.Rotation = unpacked_data[322]
        self.UserInterrupted = bool(unpacked_data[323])
        self.VerMaj = unpacked_data[324]
        self.VerMin = unpacked_data[325]

        return self


class RolloutFileRecord(dict):

    ROLLOUTCONTEXT = 0

    def __init__(self, version=-1, **kw):
        """ Create a game file record based upon the given file version
        number. The file version is first found in a HeaderMatchEntry
        object. The version needs to be propogated to all other game
        file objects within the same archive.
        """
        defaults = {
            'Name': 'RolloutFileRecord',
            'EntryType': 0,
            'Record': None,
            'Version': version
            }
        super(RolloutFileRecord, self).__init__(defaults, **kw)

    def __setattr__(self, key, value):
        self[key] = value

    def __getattr__(self, key):
       return self[key]

    def fromstream(self, stream):
        # If we are at EOF then return
        if len(stream.read(1)) <= 0:
            return None

        stream.seek(-1, _os.SEEK_CUR)
        startpos = stream.tell()

        # Using the appropriate class, read the data stream
        self.Record = RolloutContextEntry()
        self.Record.Version = self.Version
        self.Record.fromstream(stream)
        realrecsize = stream.tell() - startpos
        # Each record is actually 2184 bytes long. We need to advance past
        # the unused filler data to be at the start of the next record
        stream.seek(self.Record.SIZEOFREC - realrecsize, _os.SEEK_CUR)

        return self.Record


if __name__ == '__main__':
    pass



"""Basic tests for AnkiGammon functionality."""

import pytest

from ankigammon.models import Position, Player, CubeState, Decision, Move, DecisionType
from ankigammon.utils.xgid import parse_xgid, encode_xgid
from ankigammon.utils.ogid import parse_ogid, encode_ogid
from ankigammon.utils.gnuid import parse_gnuid, encode_gnuid
from ankigammon.utils.move_parser import MoveParser
from ankigammon.renderer.svg_board_renderer import SVGBoardRenderer
from ankigammon.parsers.xg_text_parser import XGTextParser
from ankigammon.parsers.xg_binary_parser import XGBinaryParser


class TestXGIDParsing:
    """Test XGID parsing and encoding."""

    def test_parse_basic_xgid(self):
        """Test parsing a basic XGID."""
        xgid = "XGID=---BBBBAAA---Ac-bbccbAA-A-:1:1:-1:63:4:3:0:5:8"
        position, metadata = parse_xgid(xgid)

        # Check cube
        assert metadata['cube_value'] == 2  # 2^1
        # Cube position is perspective-dependent:
        # turn=-1 (X on roll), cube_position=1 (BOTTOM in X's view) = X owns
        assert metadata['cube_owner'] == CubeState.X_OWNS

        # Check turn
        assert metadata['on_roll'] == Player.X  # -1 = X's turn

        # Check dice
        assert metadata['dice'] == (6, 3)

        # Check match
        assert metadata['match_length'] == 5

    def test_encode_xgid(self):
        """Test encoding a position to XGID."""
        position = Position()
        position.points[1] = -2  # 2 O checkers on point 1

        xgid = encode_xgid(
            position,
            cube_value=2,
            dice=(6, 3),
            match_length=7
        )

        assert "XGID=" in xgid
        assert ":1:" in xgid  # Cube value 2^1

    def test_xgid_roundtrip(self):
        """Test encoding and decoding produces same result."""
        original_xgid = "XGID=--------------------c-e-B-:0:0:1:52:0:0:0:0:0"
        position, metadata = parse_xgid(original_xgid)

        # Re-encode
        new_xgid = encode_xgid(
            position,
            cube_value=metadata['cube_value'],
            cube_owner=metadata['cube_owner'],
            dice=metadata.get('dice'),
            on_roll=metadata['on_roll'],
            score_x=metadata['score_x'],
            score_o=metadata['score_o'],
            match_length=metadata['match_length']
        )

        # Parse again
        position2, metadata2 = parse_xgid(new_xgid)

        # Compare key fields
        assert position.points == position2.points
        assert metadata['cube_value'] == metadata2['cube_value']
        assert metadata['on_roll'] == metadata2['on_roll']


class TestOGIDParsing:
    """Test OGID parsing and encoding."""

    def test_parse_position_only_ogid(self):
        """Test parsing position-only OGID (first 3 fields)."""
        ogid = "11jjjjjhhhccccc:ooddddd88866666:N0N"
        position, metadata = parse_ogid(ogid)

        # Check position - starting position
        # White/X: 2 on pt1, 5 on pt12(c), 3 on pt17(h), 5 on pt19(j)
        assert position.points[1] == 2   # 2 X checkers on point 1
        assert position.points[12] == 5  # 5 X checkers on point 12 (c)
        assert position.points[17] == 3  # 3 X checkers on point 17 (h)
        assert position.points[19] == 5  # 5 X checkers on point 19 (j)

        # Black/O: 2 on pt24(o), 5 on pt13(d), 3 on pt8, 5 on pt6
        assert position.points[24] == -2  # 2 O checkers on point 24
        assert position.points[13] == -5  # 5 O checkers on point 13 (d)
        assert position.points[8] == -3   # 3 O checkers on point 8
        assert position.points[6] == -5   # 5 O checkers on point 6

        # Check cube - neutral at 1
        assert metadata['cube_value'] == 1
        assert metadata['cube_owner'] == CubeState.CENTERED
        assert metadata['cube_action'] == 'N'

    def test_parse_full_ogid(self):
        """Test parsing full OGID with all metadata fields."""
        ogid = "11jjjjjhhhccccc:ooddddd88866666:N0N:65:W:IW:0:0:7:0"
        position, metadata = parse_ogid(ogid)

        # Check metadata
        assert metadata['cube_value'] == 1
        assert metadata['cube_owner'] == CubeState.CENTERED
        assert metadata['dice'] == (6, 5)
        assert metadata['on_roll'] == Player.X  # W = White = X
        assert metadata['game_state'] == 'IW'
        assert metadata['score_x'] == 0
        assert metadata['score_o'] == 0
        assert metadata['match_length'] == 7
        assert metadata['move_id'] == 0

    def test_parse_ogid_with_cube_doubled(self):
        """Test parsing OGID with doubled cube."""
        ogid = "jjjjkk:od88866:W2O:43:B:IW:2:1:7:15"
        position, metadata = parse_ogid(ogid)

        # Check cube - White owns at 4, offered
        assert metadata['cube_value'] == 4  # 2^2
        assert metadata['cube_owner'] == CubeState.X_OWNS  # White = X
        assert metadata['cube_action'] == 'O'  # Offered

        # Check other metadata
        assert metadata['dice'] == (4, 3)
        assert metadata['on_roll'] == Player.O  # B = Black = O
        assert metadata['score_x'] == 2
        assert metadata['score_o'] == 1
        assert metadata['match_length'] == 7
        assert metadata['move_id'] == 15

    def test_parse_ogid_crawford_game(self):
        """Test parsing OGID with Crawford game modifier."""
        ogid = "11jjjjjhhhccccc:ooddddd88866666:N0N:65:W:IW:6:5:7C:42"
        position, metadata = parse_ogid(ogid)

        assert metadata['match_length'] == 7
        assert metadata['match_modifier'] == 'C'  # Crawford
        assert metadata['score_x'] == 6
        assert metadata['score_o'] == 5
        assert metadata['move_id'] == 42

    def test_encode_position_only_ogid(self):
        """Test encoding position to OGID (position-only format)."""
        position = Position()
        # Set up starting position
        position.points[1] = 2    # X
        position.points[12] = 5   # X
        position.points[17] = 3   # X
        position.points[19] = 5   # X
        position.points[6] = -5   # O
        position.points[8] = -3   # O
        position.points[13] = -5  # O
        position.points[24] = -2  # O

        ogid = encode_ogid(position, only_position=True)

        # Should have exactly 3 fields
        parts = ogid.split(':')
        assert len(parts) == 3

        # Field 1: White/X checkers (sorted)
        assert parts[0] == "11ccccchhhjjjjj"

        # Field 2: Black/O checkers (sorted)
        assert parts[1] == "66666888dddddoo"

        # Field 3: Cube state
        assert parts[2] == "N0N"

    def test_encode_full_ogid(self):
        """Test encoding position with full metadata."""
        position = Position()
        position.points[1] = 2

        ogid = encode_ogid(
            position,
            cube_value=4,
            cube_owner=CubeState.X_OWNS,
            cube_action='T',
            dice=(6, 5),
            on_roll=Player.X,
            game_state='IW',
            score_x=2,
            score_o=1,
            match_length=7,
            move_id=10
        )

        parts = ogid.split(':')
        assert len(parts) >= 10

        # Check key fields
        assert parts[2] == "W2T"  # Cube: White owns 4 (2^2), Taken
        assert parts[3] == "65"   # Dice
        assert parts[4] == "W"    # White to move
        assert parts[5] == "IW"   # Game state
        assert parts[6] == "2"    # White score
        assert parts[7] == "1"    # Black score
        assert parts[8] == "7"    # Match length
        assert parts[9] == "10"   # Move ID

    def test_ogid_roundtrip(self):
        """Test encoding and decoding produces same result."""
        original_ogid = "11jjjjjhhhccccc:ooddddd88866666:N0N:52:W:IW:0:0:7:0"
        position, metadata = parse_ogid(original_ogid)

        # Re-encode
        new_ogid = encode_ogid(
            position,
            cube_value=metadata['cube_value'],
            cube_owner=metadata['cube_owner'],
            cube_action=metadata['cube_action'],
            dice=metadata.get('dice'),
            on_roll=metadata.get('on_roll'),
            game_state=metadata.get('game_state', ''),
            score_x=metadata.get('score_x', 0),
            score_o=metadata.get('score_o', 0),
            match_length=metadata.get('match_length'),
            move_id=metadata.get('move_id')
        )

        # Parse again
        position2, metadata2 = parse_ogid(new_ogid)

        # Compare key fields
        assert position.points == position2.points
        assert metadata['cube_value'] == metadata2['cube_value']
        assert metadata['on_roll'] == metadata2['on_roll']
        assert metadata['dice'] == metadata2['dice']

    def test_position_from_ogid(self):
        """Test Position.from_ogid() class method."""
        ogid = "11jjjjjhhhccccc:ooddddd88866666:N0N"
        position = Position.from_ogid(ogid)

        # Check starting position
        assert position.points[1] == 2
        assert position.points[19] == 5
        assert position.points[24] == -2
        assert position.points[13] == -5

    def test_position_to_ogid(self):
        """Test Position.to_ogid() instance method."""
        position = Position()
        position.points[1] = 2
        position.points[6] = -3

        ogid = position.to_ogid(
            cube_value=2,
            cube_owner=CubeState.O_OWNS,
            dice=(4, 3),
            on_roll=Player.O
        )

        # Parse back
        parsed_pos, metadata = parse_ogid(ogid)

        assert parsed_pos.points[1] == 2
        assert parsed_pos.points[6] == -3
        assert metadata['cube_value'] == 2
        assert metadata['cube_owner'] == CubeState.O_OWNS


class TestGNUIDParsing:
    """Test GNUID parsing and encoding."""

    def test_parse_starting_position(self):
        """Test parsing GNUID starting position."""
        # Starting position from GNU Backgammon manual
        gnuid = "4HPwATDgc/ABMA:MIEFAAAAAAAA"
        position, metadata = parse_gnuid(gnuid)

        # Check that we have a valid position (15 checkers per side)
        total_x = sum(count for count in position.points if count > 0)
        total_o = sum(abs(count) for count in position.points if count < 0)

        assert total_x + position.x_off == 15
        assert total_o + position.o_off == 15

        # Metadata should have cube and match info
        assert 'cube_value' in metadata
        assert 'cube_owner' in metadata

    def test_parse_position_only_gnuid(self):
        """Test parsing GNUID with position ID only (no match ID)."""
        gnuid = "4HPwATDgc/ABMA"
        position, _ = parse_gnuid(gnuid)

        # Should still parse position
        total_x = sum(count for count in position.points if count > 0)
        total_o = sum(abs(count) for count in position.points if count < 0)

        assert total_x + position.x_off == 15
        assert total_o + position.o_off == 15

    def test_parse_gnuid_with_prefix(self):
        """Test parsing GNUID with GNUID= or GNUBGID prefix."""
        gnuid1 = "GNUID=4HPwATDgc/ABMA:MIEFAAAAAAAA"
        gnuid2 = "GNUBGID 4HPwATDgc/ABMA:MIEFAAAAAAAA"

        position1, _ = parse_gnuid(gnuid1)
        position2, _ = parse_gnuid(gnuid2)

        # Both should parse to same position
        assert position1.points == position2.points

    def test_parse_gnuid_with_match_metadata(self):
        """Test parsing GNUID with full match metadata."""
        # Example with match info
        gnuid = "4HPwATDgc/ABMA:8IhuACAACAAE"
        position, metadata = parse_gnuid(gnuid)

        # Should have match metadata
        assert 'cube_value' in metadata
        assert 'on_roll' in metadata
        assert 'score_x' in metadata
        assert 'score_o' in metadata

    def test_encode_position_only_gnuid(self):
        """Test encoding position to GNUID (position-only format)."""
        position = Position()
        # Set up a simple position
        position.points[1] = 2   # X
        position.points[6] = -5  # O
        position.points[13] = -5 # O
        position.points[24] = 2  # X

        gnuid = encode_gnuid(position, on_roll=Player.X, only_position=True)

        # Should be exactly 14 characters (Position ID only)
        assert len(gnuid) == 14
        # Should be valid Base64
        assert all(c in 'ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/' for c in gnuid)

    def test_encode_full_gnuid(self):
        """Test encoding position with full metadata."""
        position = Position()
        position.points[1] = 2

        gnuid = encode_gnuid(
            position,
            cube_value=2,
            cube_owner=CubeState.X_OWNS,
            dice=(6, 5),
            on_roll=Player.X,
            score_x=2,
            score_o=1,
            match_length=7
        )

        # Should have format: PositionID:MatchID
        parts = gnuid.split(':')
        assert len(parts) == 2
        assert len(parts[0]) == 14  # Position ID
        assert len(parts[1]) == 12  # Match ID

    def test_gnuid_roundtrip(self):
        """Test encoding and decoding produces same result."""
        # Create a position with only one player's checkers (simpler case)
        position = Position()
        position.points[6] = 5
        position.points[8] = 3
        position.points[13] = 5
        position.points[24] = 2

        # Encode to GNUID
        gnuid = encode_gnuid(
            position,
            cube_value=1,
            cube_owner=CubeState.CENTERED,
            dice=(5, 2),
            on_roll=Player.X,
            score_x=0,
            score_o=0,
            match_length=7
        )

        # Parse back
        position2, metadata = parse_gnuid(gnuid)

        # Positions should match (at least structurally)
        total_x1 = sum(c for c in position.points if c > 0)
        total_x2 = sum(c for c in position2.points if c > 0)
        assert total_x1 == total_x2

        # Metadata should match
        assert metadata['cube_value'] == 1
        assert metadata['on_roll'] == Player.X
        assert metadata['dice'] == (5, 2)
        assert metadata['match_length'] == 7

    def test_position_from_gnuid(self):
        """Test Position.from_gnuid() class method."""
        gnuid = "4HPwATDgc/ABMA:MIEFAAAAAAAA"
        position = Position.from_gnuid(gnuid)

        # Check valid position
        total_x = sum(count for count in position.points if count > 0)
        total_o = sum(abs(count) for count in position.points if count < 0)

        assert total_x + position.x_off == 15
        assert total_o + position.o_off == 15

    def test_position_to_gnuid(self):
        """Test Position.to_gnuid() instance method."""
        position = Position()
        position.points[1] = 2
        position.points[6] = 3

        gnuid = position.to_gnuid(
            cube_value=2,
            cube_owner=CubeState.X_OWNS,
            dice=(4, 3),
            on_roll=Player.X
        )

        # Parse back
        parsed_pos, metadata = parse_gnuid(gnuid)

        # Verify metadata
        assert metadata['cube_value'] == 2
        assert metadata['cube_owner'] == CubeState.X_OWNS
        assert metadata['dice'] == (4, 3)

        # Verify total checkers match (position encoding may have perspective issues)
        total_checkers_orig = sum(abs(c) for c in position.points)
        total_checkers_parsed = sum(abs(c) for c in parsed_pos.points)
        assert total_checkers_orig == total_checkers_parsed

    def test_gnuid_crawford_game(self):
        """Test GNUID encoding/decoding with Crawford game."""
        position = Position()
        position.points[1] = 2

        gnuid = encode_gnuid(
            position,
            on_roll=Player.X,
            score_x=6,
            score_o=5,
            match_length=7,
            crawford=True
        )

        # Parse back
        _, metadata = parse_gnuid(gnuid)

        assert metadata['crawford'] is True
        assert metadata['score_x'] == 6
        assert metadata['score_o'] == 5
        assert metadata['match_length'] == 7


class TestMoveParser:
    """Test move notation parsing."""

    def test_parse_simple_move(self):
        """Test parsing simple move notation."""
        moves = MoveParser.parse_move_notation("13/9 6/5")
        assert len(moves) == 2
        assert moves[0] == (13, 9)
        assert moves[1] == (6, 5)

    def test_parse_bar_move(self):
        """Test parsing bar entry."""
        moves = MoveParser.parse_move_notation("bar/22")
        assert len(moves) == 1
        assert moves[0] == (0, 22)

    def test_parse_bearoff(self):
        """Test parsing bear-off move."""
        moves = MoveParser.parse_move_notation("6/off")
        assert len(moves) == 1
        assert moves[0] == (6, 26)


class TestPosition:
    """Test Position model."""

    def test_position_creation(self):
        """Test creating a position."""
        position = Position()
        assert len(position.points) == 26
        assert position.x_off == 0
        assert position.o_off == 0

    def test_position_copy(self):
        """Test copying a position."""
        position = Position()
        position.points[1] = 5
        position.x_off = 2

        copy = position.copy()
        copy.points[1] = 3
        copy.x_off = 1

        # Original should be unchanged
        assert position.points[1] == 5
        assert position.x_off == 2

    def test_from_xgid(self):
        """Test creating position from XGID."""
        xgid = "XGID=---BBBBAAA---Ac-bbccbAA-A-:1:1:-1:63:4:3:0:5:8"
        position = Position.from_xgid(xgid)

        assert len(position.points) == 26


class TestDecision:
    """Test Decision model."""

    def test_decision_creation(self):
        """Test creating a decision."""
        position = Position()
        moves = [
            Move(notation="13/9 6/5", equity=0.234, rank=1),
            Move(notation="24/20 13/9", equity=0.221, error=0.013, rank=2),
        ]

        decision = Decision(
            position=position,
            on_roll=Player.O,
            dice=(6, 3),
            candidate_moves=moves
        )

        assert decision.on_roll == Player.O
        assert len(decision.candidate_moves) == 2

    def test_get_best_move(self):
        """Test getting best move."""
        position = Position()
        moves = [
            Move(notation="13/9 6/5", equity=0.234, rank=1),
            Move(notation="24/20 13/9", equity=0.221, error=0.013, rank=2),
        ]

        decision = Decision(
            position=position,
            candidate_moves=moves
        )

        best = decision.get_best_move()
        assert best is not None
        assert best.rank == 1
        assert best.notation == "13/9 6/5"

    def test_short_display_text(self):
        """Test short display text for different game types."""
        position = Position()

        # Money game - no match length
        decision = Decision(
            position=position,
            on_roll=Player.O,
            dice=(6, 3),
            score_x=0,
            score_o=0,
            match_length=0,
            decision_type=DecisionType.CHECKER_PLAY
        )
        assert decision.get_short_display_text() == "Checker | 63 | Money"

        # Match game - with match length
        decision = Decision(
            position=position,
            on_roll=Player.O,
            dice=(5, 2),
            score_x=3,
            score_o=4,
            match_length=7,
            decision_type=DecisionType.CHECKER_PLAY
        )
        assert decision.get_short_display_text() == "Checker | 52 | 3-4 of 7"

        # Crawford game - with match length and Crawford flag
        decision = Decision(
            position=position,
            on_roll=Player.X,
            dice=(6, 6),
            score_x=1,
            score_o=4,
            match_length=5,
            crawford=True,
            decision_type=DecisionType.CHECKER_PLAY
        )
        assert decision.get_short_display_text() == "Checker | 66 | 1-4 of 5 Crawford"

        # Cube decision in Crawford game
        decision = Decision(
            position=position,
            on_roll=Player.O,
            score_x=6,
            score_o=5,
            match_length=7,
            crawford=True,
            decision_type=DecisionType.CUBE_ACTION
        )
        assert decision.get_short_display_text() == "Cube | 6-5 of 7 Crawford"


class TestSVGBoardRenderer:
    """Test SVG board renderer."""

    def test_render_svg_generates_valid_markup(self):
        """Ensure SVG renderer generates valid SVG markup."""
        renderer = SVGBoardRenderer()

        # Create a simple position
        position = Position()
        position.points[1] = -2
        position.points[24] = 2

        # Render to SVG
        svg = renderer.render_svg(position, Player.O, dice=(3, 5))

        # Check that it's valid SVG
        assert '<svg' in svg
        assert 'viewBox="0 0 880 600"' in svg
        assert '</svg>' in svg
        assert len(svg) > 5000  # Should be a reasonable size


class TestXGTextParser:
    """Test XG text parser."""

    def test_parse_cube_decision(self):
        """Test parsing cube decision analysis."""
        text = """XGID=--BBbBB-----aE----Be-c-bb-:1:-1:-1:00:0:0:0:0:8

X:Player 1   O:Player 2
Score is X:0 O:0. Unlimited Game
Cube: 2, X own cube
X on roll, cube action

Analyzed in 2-ply
Player Winning Chances:   49.42% (G:11.68% B:0.07%)
Opponent Winning Chances: 50.58% (G:10.58% B:0.22%)

Cubeless Equities: No Double=-0.002, Double=-0.004

Cubeful Equities:
       No redouble:     +0.172
       Redouble/Take:   -0.361 (-0.533)
       Redouble/Pass:   +1.000 (+0.828)

Best Cube action: No redouble / Take

eXtreme Gammon Version: 2.10"""

        decisions = XGTextParser.parse_string(text)
        assert len(decisions) == 1

        decision = decisions[0]
        assert decision.decision_type == DecisionType.CUBE_ACTION
        assert decision.cube_value == 2
        # Should have all 5 cube options
        assert len(decision.candidate_moves) == 5

        # Check best move (rank 1)
        best_move = decision.get_best_move()
        assert best_move.notation == "No Redouble/Take"
        assert best_move.equity == pytest.approx(0.172, abs=0.001)
        assert best_move.rank == 1

        # Verify all 5 options are present
        notations = [m.notation for m in decision.candidate_moves]
        assert "No Redouble/Take" in notations
        assert "Redouble/Take" in notations
        assert "Redouble/Pass" in notations
        assert "Too good/Take" in notations
        assert "Too good/Pass" in notations

    def test_parse_too_good_cube_decision(self):
        """Test parsing 'Too good' cube decision analysis."""
        text = """XGID=--BBbBB-----a-----Be-c-bbE:1:-1:-1:00:0:0:0:0:8

X:Player 1   O:Player 2
Score is X:0 O:0. Unlimited Game
Cube: 2, X own cube
X on roll, cube action

Analyzed in 2-ply
Player Winning Chances:   88.42% (G:81.23% B:13.70%)
Opponent Winning Chances: 11.58% (G:1.09% B:0.02%)

Cubeless Equities: No Double=+1.707, Double=+3.413

Cubeful Equities:
       No redouble:     +1.761
       Redouble/Take:   +3.332 (+1.570)
       Redouble/Pass:   +1.000 (-0.761)

Best Cube action: Too good to redouble / Pass

eXtreme Gammon Version: 2.10"""

        decisions = XGTextParser.parse_string(text)
        assert len(decisions) == 1

        decision = decisions[0]
        assert decision.decision_type == DecisionType.CUBE_ACTION
        assert decision.cube_value == 2
        assert len(decision.candidate_moves) == 5

        # Check best move is "Too good/Pass"
        best_move = decision.get_best_move()
        assert best_move.notation == "Too good/Pass"
        assert best_move.rank == 1

        # Verify all 5 options are present with "Redouble" terminology
        notations = [m.notation for m in decision.candidate_moves]
        assert "No Redouble/Take" in notations
        assert "Redouble/Take" in notations
        assert "Redouble/Pass" in notations
        assert "Too good/Take" in notations
        assert "Too good/Pass" in notations

    def test_parse_checker_play(self):
        """Test parsing checker play analysis."""
        text = """XGID=--BBbBB-----aE----Be-c-bb-:1:-1:-1:54:0:0:0:0:8

X:Player 1   O:Player 2
Score is X:0 O:0. Unlimited Game
Cube: 2, X own cube
X to play 54

    1. 3-ply       13/4                         eq:+0.165
      Player:   49.07% (G:13.31% B:0.11%)
      Opponent: 50.93% (G:10.59% B:0.24%)

    2. 3-ply       13/9 6/1                     eq:+0.026 (-0.140)
      Player:   44.31% (G:10.20% B:0.06%)
      Opponent: 55.69% (G:10.59% B:0.24%)

eXtreme Gammon Version: 2.10"""

        decisions = XGTextParser.parse_string(text)
        assert len(decisions) == 1

        decision = decisions[0]
        assert decision.decision_type == DecisionType.CHECKER_PLAY
        assert decision.dice == (5, 4)
        assert len(decision.candidate_moves) == 2

        # Check moves
        assert decision.candidate_moves[0].notation == "13/4"
        assert decision.candidate_moves[0].equity == pytest.approx(0.165, abs=0.001)

        assert decision.candidate_moves[1].notation == "13/9 6/1"
        assert decision.candidate_moves[1].error == pytest.approx(0.140, abs=0.001)

    def test_parse_crawford_game(self):
        """Test parsing Crawford game match info."""
        text = """XGID=-b---CC-C---eD---c-e---AA-:0:0:1:66:4:1:1:5:10

X:Player 1   O:Player 2
Score is X:1 O:4 5 pt.(s) match.
 +13-14-15-16-17-18------19-20-21-22-23-24-+
 | X           O    |   | O           X  X |
 | X           O    |   | O                |
 | X           O    |   | O                |
 | X                |   | O                |
 |                  |   | O                |
 |                  |BAR|                  |
 | O                |   |                  |
 | O                |   |                  |
 | O           X    |   | X  X             |
 | O           X    |   | X  X           O |
 | O           X    |   | X  X           O |
 +12-11-10--9--8--7-------6--5--4--3--2--1-+
Pip count  X: 156  O: 167 X-O: 1-4/5 Crawford
Cube: 1
X to play 66

    1. 2-ply       13/7(4)                      eq:+0.590
      Player:   67.47% (G:23.33% B:1.90%)
      Opponent: 32.53% (G:6.44% B:0.15%)

    2. 2-ply       13/7(2) 8/2(2)               eq:+0.578 (-0.012)
      Player:   65.80% (G:25.43% B:1.84%)
      Opponent: 34.20% (G:7.40% B:0.28%)

eXtreme Gammon Version: 2.10, MET: Kazaross XG2"""

        decisions = XGTextParser.parse_string(text)
        assert len(decisions) == 1

        decision = decisions[0]
        # Check Crawford flag is set
        assert decision.crawford is True
        # Check match length
        assert decision.match_length == 5
        # Check scores
        assert decision.score_x == 1
        assert decision.score_o == 4
        # Check metadata text includes Crawford indicator
        metadata_text = decision.get_metadata_text()
        assert "Crawford" in metadata_text
        assert "5pt" in metadata_text


class TestXGBinaryParser:
    """Test XG binary parser move notation conversion."""

    def test_compound_move_with_intermediate_hit(self):
        """
        Test that hits at intermediate points are preserved in compound moves.

        When a move involves multiple sub-moves (e.g., 20/15 then 15/5),
        and there's a hit at an intermediate point (e.g., at point 15),
        the notation should preserve the hit marker at that intermediate point.

        Expected: "20/15* 15/5 13/8" (shows hit at point 15)
        Not: "20/5 13/8" (loses the hit marker)
        """
        # Create a position with an opponent blot at point 15
        position = Position()
        # Clear all checkers first
        for i in range(26):
            position.points[i] = 0

        # Set up position: Player O on roll (bottom player)
        # Place O checker at point 20 (will move from here)
        position.points[20] = -1  # O's checker (negative)

        # Place X checker at point 15 (will be hit)
        position.points[15] = 1   # X's blot (positive)

        # Place O checker at point 13
        position.points[13] = -1  # O's checker

        # XG move notation (0-based): 19/14 14/4 12/7
        # Standard notation (1-based): 20/15* 15/5 13/8
        # The first move 20/15 should hit the X blot at point 15
        xg_moves = (19, 14, 14, 4, 12, 7, -1, -1)

        # Convert move notation
        notation = XGBinaryParser._convert_move_notation(
            xg_moves,
            position=position,
            on_roll=Player.O
        )

        # Verify the hit marker is present
        assert "*" in notation, "Hit marker should be present in the notation"

        # Verify the hit is at point 15 (intermediate point)
        assert "20/15*" in notation or "15*" in notation, \
            "Hit should be shown at the intermediate point (15)"

        # Verify the move is NOT fully combined into "20/5"
        # The compound move should be split at the hit point
        if "20/5" in notation:
            assert "20/15*" in notation, \
                "If showing 20/5, must also show 20/15* for the hit"

        # The expected notation should be "20/15* 15/5 13/8" or similar
        # (order may vary, but hit at 15 must be shown)
        assert notation == "20/15* 15/5 13/8" or \
               notation == "20/15* 13/8 15/5" or \
               notation == "13/8 20/15* 15/5", \
            f"Expected notation with hit at point 15, got: {notation}"

    def test_single_hit_on_same_destination(self):
        """
        Test that only the first checker to land on a point shows a hit marker.

        When two checkers land on the same destination and there's a blot,
        only the first checker hits. The second lands on an empty point.

        Expected: "6/1* 4/1" (one hit marker)
        Not: "6/1* 4/1*" (two hit markers)
        """
        # Create a position with opponent blot at point 1
        position = Position()
        for i in range(26):
            position.points[i] = 0

        # Player X on roll
        # In XG binary positions, the player on roll has negative checkers,
        # opponent has positive checkers
        position.points[6] = -1  # X's checker at point 6
        position.points[4] = -1  # X's checker at point 4
        position.points[1] = 1   # O's blot at point 1 (will be hit)

        # XG move notation (0-based): 5/0 3/0
        # Standard notation (1-based): 6/1* 4/1
        # Only the FIRST checker (6/1) should show hit marker
        xg_moves = (5, 0, 3, 0, -1, -1, -1, -1)

        notation = XGBinaryParser._convert_move_notation(
            xg_moves,
            position=position,
            on_roll=Player.X
        )

        # Check that exactly one hit marker exists
        hit_count = notation.count('*')
        assert hit_count == 1, f"Expected 1 hit marker, found {hit_count} in: {notation}"

        # Check that 6/1 has the hit marker
        assert "6/1*" in notation, f"First move (6/1) should have hit marker in: {notation}"

        # Check that 4/1 does NOT have hit marker
        assert "4/1*" not in notation, f"Second move (4/1) should not have hit marker in: {notation}"
        assert "4/1" in notation, f"Second move (4/1) should be present without hit marker in: {notation}"



"""
Application entry point for GUI mode.
"""

import sys
import os
from pathlib import Path
from PySide6.QtWidgets import QApplication, QSplashScreen, QGraphicsDropShadowEffect
from PySide6.QtCore import Qt, QTimer, QRectF
from PySide6.QtGui import QIcon, QPixmap, QPainter, QColor, QFont, QLinearGradient, QPainterPath, QPen

from ankigammon.gui.main_window import MainWindow
from ankigammon.gui.resources import get_resource_path
from ankigammon.settings import get_settings


def set_windows_app_id():
    """
    Set Windows AppUserModelID to display custom icon in taskbar.

    On Windows, this enables the application's icon to appear in the taskbar
    instead of the default Python icon.
    """
    if sys.platform == 'win32':
        try:
            import ctypes
            ctypes.windll.shell32.SetCurrentProcessExplicitAppUserModelID('AnkiGammon.GUI.1.0')
        except Exception:
            pass


def create_splash_screen(icon_path: Path) -> QSplashScreen:
    """
    Create a splash screen with modern styling.

    Features rounded corners, gradient background, branded border,
    drop shadow, and antialiased rendering.

    Args:
        icon_path: Path to the application icon

    Returns:
        QSplashScreen: Configured splash screen
    """
    # Create pixmap with margin for drop shadow
    splash_pix = QPixmap(440, 340)
    splash_pix.fill(Qt.transparent)

    # Initialize painter with antialiasing
    painter = QPainter(splash_pix)
    painter.setRenderHint(QPainter.Antialiasing)
    painter.setRenderHint(QPainter.SmoothPixmapTransform)
    painter.setRenderHint(QPainter.TextAntialiasing)

    # Define rounded rectangle with 20px margin for shadow
    rect = QRectF(20, 20, 400, 300)
    corner_radius = 12

    path = QPainterPath()
    path.addRoundedRect(rect, corner_radius, corner_radius)

    # Apply gradient background
    gradient = QLinearGradient(20, 20, 20, 320)
    gradient.setColorAt(0.0, QColor("#1e1e2e"))
    gradient.setColorAt(0.3, QColor("#262637"))
    gradient.setColorAt(0.7, QColor("#262637"))
    gradient.setColorAt(1.0, QColor("#1e1e2e"))
    painter.fillPath(path, gradient)

    # Load and draw application icon
    if icon_path.exists():
        icon_pixmap = QPixmap(str(icon_path))
        icon_size = 120
        scaled_icon = icon_pixmap.scaled(icon_size, icon_size, Qt.KeepAspectRatio, Qt.SmoothTransformation)

        icon_x = 20 + (400 - scaled_icon.width()) // 2
        icon_y = 20 + 40

        painter.drawPixmap(icon_x, icon_y, scaled_icon)

    # Draw border
    border_pen = QPen(QColor("#89b4fa"), 2.5)
    border_pen.setJoinStyle(Qt.RoundJoin)
    border_pen.setCapStyle(Qt.RoundCap)
    painter.setPen(border_pen)
    painter.setBrush(Qt.NoBrush)
    painter.drawPath(path)

    # Draw application name
    painter.setPen(QColor("#f5e0dc"))
    title_font = QFont("Segoe UI", 24, QFont.Bold)
    title_font.setLetterSpacing(QFont.AbsoluteSpacing, 0.5)
    painter.setFont(title_font)
    painter.drawText(20, 20 + 170, 400, 40, Qt.AlignCenter, "AnkiGammon")

    # Draw tagline
    painter.setPen(QColor("#b4befe"))
    tagline_font = QFont("Segoe UI", 10)
    painter.setFont(tagline_font)
    painter.drawText(20, 20 + 215, 400, 20, Qt.AlignCenter, "Backgammon Analysis to Flashcards")

    # Draw loading indicator
    painter.setPen(QColor("#a6adc8"))
    loading_font = QFont("Segoe UI", 11)
    painter.setFont(loading_font)
    painter.drawText(20, 20 + 240, 400, 30, Qt.AlignCenter, "Loading...")

    painter.end()

    # Create frameless splash screen
    splash = QSplashScreen(splash_pix, Qt.FramelessWindowHint | Qt.WindowStaysOnTopHint)

    # Add drop shadow effect
    shadow = QGraphicsDropShadowEffect()
    shadow.setBlurRadius(30)
    shadow.setColor(QColor(0, 0, 0, 140))
    shadow.setOffset(0, 6)
    splash.setGraphicsEffect(shadow)

    return splash


def main():
    """
    Launch the GUI application.

    Returns:
        int: Application exit code
    """
    import multiprocessing
    import logging

    # CRITICAL: Required for PyInstaller + multiprocessing on Windows
    # Without this, worker processes will spawn new GUI windows when using ProcessPoolExecutor
    multiprocessing.freeze_support()

    logging.basicConfig(
        level=logging.INFO,
        format='%(asctime)s - %(name)s - %(levelname)s - %(message)s'
    )

    set_windows_app_id()

    # Enable High DPI scaling
    QApplication.setHighDpiScaleFactorRoundingPolicy(
        Qt.HighDpiScaleFactorRoundingPolicy.PassThrough
    )
    QApplication.setAttribute(Qt.AA_EnableHighDpiScaling)
    QApplication.setAttribute(Qt.AA_UseHighDpiPixmaps)

    app = QApplication(sys.argv)
    app.setApplicationName("AnkiGammon")
    app.setOrganizationName("AnkiGammon")
    app.setOrganizationDomain("github.com/Deinonychus999/AnkiGammon")

    # Use Fusion style for consistent cross-platform appearance
    app.setStyle('Fusion')

    # Set platform-specific application icon
    # macOS: .icns, Windows: .ico, Linux: .png
    if sys.platform == 'darwin':
        icon_file = "ankigammon/gui/resources/icon.icns"
    elif sys.platform == 'win32':
        icon_file = "ankigammon/gui/resources/icon.ico"
    else:
        icon_file = "ankigammon/gui/resources/icon.png"

    icon_path = get_resource_path(icon_file)
    if icon_path.exists():
        app.setWindowIcon(QIcon(str(icon_path)))

    # Show splash screen with PNG icon for high-quality rendering
    splash_icon_path = get_resource_path("ankigammon/gui/resources/icon.png")
    splash = create_splash_screen(splash_icon_path)
    splash.show()
    app.processEvents()

    # Load and apply stylesheet
    style_path = get_resource_path("ankigammon/gui/resources/style.qss")
    if style_path.exists():
        with open(style_path, encoding='utf-8') as f:
            app.setStyleSheet(f.read())

    settings = get_settings()
    window = MainWindow(settings)

    # Display splash screen for minimum 1 second before showing main window
    def show_main_window():
        splash.finish(window)
        window.show()

    QTimer.singleShot(1000, show_main_window)

    return app.exec()


if __name__ == '__main__':
    raise SystemExit(main())



"""AnkiGammon: Convert eXtreme Gammon analysis into Anki flashcards."""

__version__ = "1.0.10"

from ankigammon.models import Decision, Move, Position, CubeState

__all__ = ["Decision", "Move", "Position", "CubeState"]



"""Analysis module for AnkiGammon."""

from ankigammon.analysis.score_matrix import (
    ScoreMatrixCell,
    generate_score_matrix,
    format_matrix_as_html
)

__all__ = [
    'ScoreMatrixCell',
    'generate_score_matrix',
    'format_matrix_as_html'
]



"""
Settings and configuration management for AnkiGammon.

Handles loading and saving user preferences such as color scheme selection.
"""

import json
import os
from pathlib import Path
from typing import Optional


class Settings:
    """Manages application settings with persistence."""

    DEFAULT_SETTINGS = {
        "default_color_scheme": "classic",
        "deck_name": "My AnkiGammon Deck",
        "show_options": True,
        "interactive_moves": True,
        "export_method": "ankiconnect",
        "gnubg_path": None,
        "gnubg_analysis_ply": 3,
        "generate_score_matrix": False,
        "board_orientation": "counter-clockwise",
        "last_apkg_directory": None,
        "import_checker_error_threshold": 0.080,
        "import_cube_error_threshold": 0.080,
        "import_include_player_x": True,
        "import_include_player_o": True,
        "import_selected_player_names": [],
        "max_mcq_options": 5,
        "check_for_updates": True,
        "last_update_check": None,
        "snooze_update_until": None,
    }

    def __init__(self, config_path: Optional[Path] = None):
        """
        Initialize settings manager.

        Args:
            config_path: Path to config file. If None, uses default location.
        """
        if config_path is None:
            config_dir = Path.home() / ".ankigammon"
            config_dir.mkdir(parents=True, exist_ok=True)
            config_path = config_dir / "config.json"

        self.config_path = config_path
        self._settings = self._load()

    def _load(self) -> dict:
        """Load settings from config file."""
        if not self.config_path.exists():
            return self.DEFAULT_SETTINGS.copy()

        try:
            with open(self.config_path, 'r', encoding='utf-8') as f:
                loaded = json.load(f)
                # Merge with defaults to ensure new settings have default values
                settings = self.DEFAULT_SETTINGS.copy()
                settings.update(loaded)
                return settings
        except (json.JSONDecodeError, IOError):
            # Use defaults if file is corrupted or unreadable
            return self.DEFAULT_SETTINGS.copy()

    def _save(self) -> None:
        """Save settings to config file."""
        try:
            with open(self.config_path, 'w', encoding='utf-8') as f:
                json.dump(self._settings, f, indent=2)
        except IOError:
            # Silently fail if unable to save
            pass

    def get(self, key: str, default=None):
        """Get a setting value."""
        return self._settings.get(key, default)

    def set(self, key: str, value) -> None:
        """Set a setting value and save to disk."""
        self._settings[key] = value
        self._save()

    @property
    def color_scheme(self) -> str:
        """Get the default color scheme."""
        return self._settings.get("default_color_scheme", "classic")

    @color_scheme.setter
    def color_scheme(self, value: str) -> None:
        """Set the default color scheme."""
        self.set("default_color_scheme", value)

    @property
    def deck_name(self) -> str:
        """Get the default deck name."""
        return self._settings.get("deck_name", "My AnkiGammon Deck")

    @deck_name.setter
    def deck_name(self, value: str) -> None:
        """Set the default deck name."""
        self.set("deck_name", value)

    @property
    def show_options(self) -> bool:
        """Get whether to show options on cards."""
        return self._settings.get("show_options", True)

    @show_options.setter
    def show_options(self, value: bool) -> None:
        """Set whether to show options on cards."""
        self.set("show_options", value)

    @property
    def interactive_moves(self) -> bool:
        """Get whether to enable interactive move visualization."""
        return self._settings.get("interactive_moves", True)

    @interactive_moves.setter
    def interactive_moves(self, value: bool) -> None:
        """Set whether to enable interactive move visualization."""
        self.set("interactive_moves", value)

    @property
    def export_method(self) -> str:
        """Get the default export method."""
        return self._settings.get("export_method", "ankiconnect")

    @export_method.setter
    def export_method(self, value: str) -> None:
        """Set the default export method."""
        self.set("export_method", value)

    @property
    def gnubg_path(self) -> Optional[str]:
        """Get the GnuBG executable path."""
        return self._settings.get("gnubg_path", None)

    @gnubg_path.setter
    def gnubg_path(self, value: Optional[str]) -> None:
        """Set the GnuBG executable path."""
        self.set("gnubg_path", value)

    @property
    def gnubg_analysis_ply(self) -> int:
        """Get the GnuBG analysis depth (ply)."""
        return self._settings.get("gnubg_analysis_ply", 3)

    @gnubg_analysis_ply.setter
    def gnubg_analysis_ply(self, value: int) -> None:
        """Set the GnuBG analysis depth (ply)."""
        self.set("gnubg_analysis_ply", value)

    @property
    def generate_score_matrix(self) -> bool:
        """Get whether to generate score matrix for cube decisions."""
        return self._settings.get("generate_score_matrix", False)

    @generate_score_matrix.setter
    def generate_score_matrix(self, value: bool) -> None:
        """Set whether to generate score matrix for cube decisions."""
        self.set("generate_score_matrix", value)

    @property
    def board_orientation(self) -> str:
        """Get the board orientation (clockwise or counter-clockwise)."""
        return self._settings.get("board_orientation", "counter-clockwise")

    @board_orientation.setter
    def board_orientation(self, value: str) -> None:
        """Set the board orientation (clockwise or counter-clockwise)."""
        if value not in ["clockwise", "counter-clockwise"]:
            raise ValueError("board_orientation must be 'clockwise' or 'counter-clockwise'")
        self.set("board_orientation", value)

    @property
    def last_apkg_directory(self) -> Optional[str]:
        """Get the last directory used for APKG export."""
        return self._settings.get("last_apkg_directory", None)

    @last_apkg_directory.setter
    def last_apkg_directory(self, value: Optional[str]) -> None:
        """Set the last directory used for APKG export."""
        self.set("last_apkg_directory", value)

    @property
    def import_checker_error_threshold(self) -> float:
        """Get the error threshold for checker play decisions."""
        return self._settings.get("import_checker_error_threshold", 0.080)

    @import_checker_error_threshold.setter
    def import_checker_error_threshold(self, value: float) -> None:
        """Set the error threshold for checker play decisions."""
        self.set("import_checker_error_threshold", value)

    @property
    def import_cube_error_threshold(self) -> float:
        """Get the error threshold for cube decisions."""
        return self._settings.get("import_cube_error_threshold", 0.080)

    @import_cube_error_threshold.setter
    def import_cube_error_threshold(self, value: float) -> None:
        """Set the error threshold for cube decisions."""
        self.set("import_cube_error_threshold", value)

    @property
    def import_include_player_x(self) -> bool:
        """Get whether to include Player X mistakes in imports."""
        return self._settings.get("import_include_player_x", True)

    @import_include_player_x.setter
    def import_include_player_x(self, value: bool) -> None:
        """Set whether to include Player X mistakes in imports."""
        self.set("import_include_player_x", value)

    @property
    def import_include_player_o(self) -> bool:
        """Get whether to include Player O mistakes in imports."""
        return self._settings.get("import_include_player_o", True)

    @import_include_player_o.setter
    def import_include_player_o(self, value: bool) -> None:
        """Set whether to include Player O mistakes in imports."""
        self.set("import_include_player_o", value)

    @property
    def import_selected_player_names(self) -> list[str]:
        """Get the list of previously selected player names."""
        return self._settings.get("import_selected_player_names", [])

    @import_selected_player_names.setter
    def import_selected_player_names(self, value: list[str]) -> None:
        """Set the list of previously selected player names."""
        if not isinstance(value, list):
            value = []

        validated = []
        seen = set()
        for name in value:
            if not isinstance(name, str):
                continue

            trimmed = name.strip()
            if not trimmed:
                continue

            # Case-insensitive deduplication
            key = trimmed.lower()
            if key not in seen:
                seen.add(key)
                validated.append(trimmed)

        self.set("import_selected_player_names", validated)

    @property
    def max_mcq_options(self) -> int:
        """Get the maximum number of MCQ options to display."""
        return self._settings.get("max_mcq_options", 5)

    @max_mcq_options.setter
    def max_mcq_options(self, value: int) -> None:
        """Set the maximum number of MCQ options to display."""
        if value < 2 or value > 10:
            raise ValueError("max_mcq_options must be between 2 and 10")
        self.set("max_mcq_options", value)

    @property
    def check_for_updates(self) -> bool:
        """Get whether to check for updates on startup."""
        return self._settings.get("check_for_updates", True)

    @check_for_updates.setter
    def check_for_updates(self, value: bool) -> None:
        """Set whether to check for updates on startup."""
        self.set("check_for_updates", value)

    @property
    def last_update_check(self) -> Optional[str]:
        """Get the timestamp of the last update check (ISO format)."""
        return self._settings.get("last_update_check", None)

    @last_update_check.setter
    def last_update_check(self, value: Optional[str]) -> None:
        """Set the timestamp of the last update check (ISO format)."""
        self.set("last_update_check", value)

    @property
    def snooze_update_until(self) -> Optional[str]:
        """Get the timestamp to snooze update notifications until (ISO format)."""
        return self._settings.get("snooze_update_until", None)

    @snooze_update_until.setter
    def snooze_update_until(self, value: Optional[str]) -> None:
        """Set the timestamp to snooze update notifications until (ISO format)."""
        self.set("snooze_update_until", value)

    def is_gnubg_available(self) -> bool:
        """
        Check if GnuBG is configured and accessible.

        Returns:
            True if gnubg_path is set and the file exists and is executable.
        """
        path = self.gnubg_path
        if path is None:
            return False
        try:
            path_obj = Path(path)
            return path_obj.exists() and os.access(path, os.X_OK)
        except (OSError, ValueError):
            return False


# Global settings instance
_settings = None


def get_settings() -> Settings:
    """Get the global settings instance."""
    global _settings
    if _settings is None:
        _settings = Settings()
    return _settings



"""Helper for generating GSAP-based checker movement animations."""

import re
from typing import List, Tuple, Dict, Optional
from ankigammon.models import Position, Move, Player
from ankigammon.utils.move_parser import MoveParser


class AnimationHelper:
    """Generates animation data and JavaScript for checker movements."""

    @staticmethod
    def parse_move_notation(notation: str, on_roll: Player) -> List[Tuple[int, int]]:
        """
        Parse backgammon move notation into (from_point, to_point) pairs.

        Args:
            notation: Move notation (e.g., "24/23 24/23" or "bar/24")
            on_roll: Player making the move

        Returns:
            List of (from_point, to_point) tuples (0-25, where 0=X bar, 25=O bar)
        """
        moves = []

        if not notation or notation == "Can't move":
            return moves

        parts = notation.strip().split()

        for part in parts:
            if '/' not in part:
                continue

            # Handle repetition notation like "6/4(4)"
            repetition_count = 1
            repetition_match = re.search(r'\((\d+)\)$', part)
            if repetition_match:
                repetition_count = int(repetition_match.group(1))
                part = re.sub(r'\(\d+\)$', '', part)

            # Handle compound notation like "6/5*/3"
            segments = part.split('/')
            segments = [seg.rstrip('*') for seg in segments]
            if len(segments) > 2:
                # Convert compound notation to individual moves
                for i in range(len(segments) - 1):
                    from_str = segments[i]
                    to_str = segments[i + 1]
                    if from_str.lower() == 'bar':
                        from_point = 0 if on_roll == Player.X else 25
                    elif from_str.lower() == 'off':
                        continue
                    else:
                        try:
                            from_point = int(from_str)
                        except ValueError:
                            continue

                    if to_str.lower() == 'bar':
                        to_point = 25 if on_roll == Player.X else 0
                    elif to_str.lower() == 'off':
                        to_point = -1
                    else:
                        try:
                            to_point = int(to_str)
                        except ValueError:
                            continue

                    for _ in range(repetition_count):
                        moves.append((from_point, to_point))
            else:
                from_str = segments[0]
                to_str = segments[1] if len(segments) > 1 else ''

                if not to_str:
                    continue

                if from_str.lower() == 'bar':
                    from_point = 0 if on_roll == Player.X else 25
                elif from_str.lower() == 'off':
                    continue
                else:
                    try:
                        from_point = int(from_str)
                    except ValueError:
                        continue

                if to_str.lower() == 'bar':
                    to_point = 25 if on_roll == Player.X else 0
                elif to_str.lower() == 'off':
                    to_point = -1
                else:
                    try:
                        to_point = int(to_str)
                    except ValueError:
                        continue

                for _ in range(repetition_count):
                    moves.append((from_point, to_point))

        return moves

    @staticmethod
    def generate_animation_javascript(
        move_notation: str,
        on_roll: Player,
        duration: float = 0.8,
        stagger: float = 0.05
    ) -> str:
        """
        Generate JavaScript code for animating a backgammon move using GSAP.

        Args:
            move_notation: Move notation string
            on_roll: Player making the move
            duration: Animation duration in seconds
            stagger: Stagger time between checkers in seconds

        Returns:
            JavaScript code string
        """
        moves = AnimationHelper.parse_move_notation(move_notation, on_roll)

        if not moves:
            return ""

        js_code = f"""
<script src="https://cdn.jsdelivr.net/npm/gsap@3/dist/gsap.min.js"></script>
<script src="https://cdn.jsdelivr.net/npm/gsap@3/dist/MotionPathPlugin.min.js"></script>
<script>
if (typeof gsap !== 'undefined') {{
    gsap.registerPlugin(MotionPathPlugin);

    const moves = {moves};
    const duration = {duration};
    const stagger = {stagger};

    function animateMove() {{
        moves.forEach((move, index) => {{
            const [fromPoint, toPoint] = move;
            const checkers = document.querySelectorAll('.checker[data-point="' + fromPoint + '"]');

            if (checkers.length > 0) {{
                const checker = checkers[0];

                gsap.to(checker, {{
                    duration: duration,
                    delay: index * stagger,
                    attr: {{
                        cx: '+=100',
                        cy: '+=0'
                    }},
                    ease: 'power2.inOut'
                }});
            }}
        }});
    }}
}}
</script>
"""
        return js_code

    @staticmethod
    def get_gsap_cdn_scripts() -> str:
        """
        Get GSAP CDN script tags.

        Returns:
            HTML script tags for GSAP and MotionPathPlugin
        """
        return """
<script src="https://cdn.jsdelivr.net/npm/gsap@3/dist/gsap.min.js"></script>
<script src="https://cdn.jsdelivr.net/npm/gsap@3/dist/MotionPathPlugin.min.js"></script>
"""

    @staticmethod
    def generate_animation_button() -> str:
        """
        Generate HTML for an animation trigger button.

        Returns:
            HTML button element
        """
        return """
<div style="text-align: center; margin: 10px 0;">
    <button onclick="animateCheckers()" class="toggle-btn">
        ▶ Animate Move
    </button>
</div>
"""



"""XGID format parsing and encoding.

XGID (eXtreme Gammon ID) is a compact text representation of a backgammon position.

Format: XGID=PPPPPPPPPPPPPPPPPPPPPPPPPP:CV:CP:T:D:S1:S2:CJ:ML:MC

Fields:
1. Position (26 chars):
   - Char 0: bar for TOP player
   - Chars 1-24: points 1-24 (from BOTTOM player's perspective)
   - Char 25: bar for BOTTOM player
   - 'A'-'P': BOTTOM player's checkers (1-16)
   - 'a'-'p': TOP player's checkers (1-16)
   - '-': empty point

2. Cube Value (CV): 2^CV (0=1, 1=2, 2=4, etc.)

3. Cube Position (CP):
   - 1: owned by BOTTOM player
   - 0: centered
   - -1: owned by TOP player

4. Turn (T):
   - 1: BOTTOM player's turn
   - -1: TOP player's turn

5. Dice (D):
   - 00: player to roll or double
   - D: player doubled, opponent must take/drop
   - B: player doubled, opponent beavered
   - R: doubled, beavered, and raccooned
   - xx: rolled dice (e.g., 63, 35, 11)

6. Score 1 (S1): BOTTOM player's score
7. Score 2 (S2): TOP player's score
8. Crawford/Jacoby (CJ): Crawford rule (match) or Jacoby rule (money)
9. Match Length (ML): 0 for money games
10. Max Cube (MC): Maximum cube value (2^MC)

Note: In our internal model, we use O for BOTTOM player and X for TOP player.
"""

import re
from typing import Optional, Tuple

from ankigammon.models import Position, Player, CubeState


def parse_xgid(xgid: str) -> Tuple[Position, dict]:
    """
    Parse an XGID string into a Position and metadata.

    Args:
        xgid: XGID string (e.g., "XGID=-a-B--C-dE---eE---c-e----B-:1:0:1:63:0:0:0:0:10")

    Returns:
        Tuple of (Position, metadata_dict)
    """
    # Remove "XGID=" prefix if present
    if xgid.startswith("XGID="):
        xgid = xgid[5:]

    # Split into components
    parts = xgid.split(':')
    if len(parts) < 9:
        raise ValueError(f"Invalid XGID format: expected 9+ parts, got {len(parts)}")

    position_str = parts[0]
    cube_value_log = int(parts[1])
    cube_position = int(parts[2])
    turn = int(parts[3])
    dice_str = parts[4]
    score_bottom = int(parts[5])
    score_top = int(parts[6])
    crawford_jacoby = int(parts[7]) if len(parts) > 7 else 0
    match_length = int(parts[8]) if len(parts) > 8 else 0
    max_cube = int(parts[9]) if len(parts) > 9 else 8

    # Parse position (encoding is perspective-dependent based on turn)
    position = _parse_position_string(position_str, turn)

    # Parse metadata
    metadata = {}

    # Cube value (2^cube_value_log)
    cube_value = 2 ** cube_value_log if cube_value_log >= 0 else 1
    metadata['cube_value'] = cube_value

    # Cube owner is perspective-dependent (flipped when X is on roll)
    if cube_position == 0:
        cube_state = CubeState.CENTERED
    elif cube_position == -1:
        cube_state = CubeState.X_OWNS if turn == 1 else CubeState.O_OWNS
    else:  # cube_position == 1
        cube_state = CubeState.O_OWNS if turn == 1 else CubeState.X_OWNS
    metadata['cube_owner'] = cube_state

    # Turn: 1 = BOTTOM player (O), -1 = TOP player (X)
    on_roll = Player.O if turn == 1 else Player.X
    metadata['on_roll'] = on_roll

    # Dice
    dice_str = dice_str.upper().strip()
    if dice_str == '00':
        # Player to roll or double (no dice shown)
        pass
    elif dice_str in ['D', 'B', 'R']:
        # Cube action pending
        metadata['decision_type'] = 'cube_action'
    elif len(dice_str) == 2 and dice_str.isdigit():
        # Rolled dice
        d1 = int(dice_str[0])
        d2 = int(dice_str[1])
        if 1 <= d1 <= 6 and 1 <= d2 <= 6:
            metadata['dice'] = (d1, d2)

    # Score: in XGID, field 5 is bottom player, field 6 is top player
    # We map bottom=O, top=X
    metadata['score_o'] = score_bottom
    metadata['score_x'] = score_top

    # Match length
    metadata['match_length'] = match_length

    # Crawford/Jacoby
    metadata['crawford_jacoby'] = crawford_jacoby

    # Max cube
    metadata['max_cube'] = 2 ** max_cube if max_cube >= 0 else 256

    return position, metadata


def _parse_position_string(pos_str: str, turn: int) -> Position:
    """
    Parse the position encoding part of XGID.

    The 26-character position string is encoded from the perspective of the player on roll:

    When turn=1 (O on roll):
    - Char 0: X's bar, Chars 1-24: points 1-24, Char 25: O's bar
    - lowercase = X checkers, uppercase = O checkers

    When turn=-1 (X on roll):
    - Char 0: O's bar, Chars 1-24: points in reverse (24-1), Char 25: X's bar
    - lowercase = X checkers, uppercase = O checkers

    Internal model (always consistent):
    - points[0] = X's bar (TOP player)
    - points[1-24] = board points (1 = O's home, 24 = X's home)
    - points[25] = O's bar (BOTTOM player)
    """
    if len(pos_str) != 26:
        raise ValueError(f"Position string must be 26 characters, got {len(pos_str)}")

    position = Position()

    if turn == 1:
        # O on roll: standard perspective
        position.points[0] = _decode_checker_count(pos_str[0], turn)
        position.points[25] = _decode_checker_count(pos_str[25], turn)

        for i in range(1, 25):
            position.points[i] = _decode_checker_count(pos_str[i], turn)
    else:
        # X on roll: flipped perspective - bars and points are reversed
        position.points[0] = _decode_checker_count(pos_str[25], turn)  # X's bar
        position.points[25] = _decode_checker_count(pos_str[0], turn)  # O's bar

        # Board points: reverse mapping
        for i in range(1, 25):
            position.points[i] = _decode_checker_count(pos_str[25 - i], turn)

    # Calculate borne-off checkers (each player starts with 15)
    total_x = sum(count for count in position.points if count > 0)
    total_o = sum(abs(count) for count in position.points if count < 0)

    position.x_off = 15 - total_x
    position.o_off = 15 - total_o

    return position


def _decode_checker_count(char: str, turn: int) -> int:
    """
    Decode a single character to checker count.

    The uppercase/lowercase mapping depends on whose turn it is:

    When turn=1 (O on roll):
    - lowercase = X checkers (positive)
    - uppercase = O checkers (negative)

    When turn=-1 (X on roll):
    - lowercase = O checkers (negative)
    - uppercase = X checkers (positive)

    Args:
        char: The character to decode
        turn: 1 if O on roll, -1 if X on roll

    Returns:
        Checker count (positive for X, negative for O, 0 for empty)
    """
    if char == '-':
        return 0

    count = 0
    if 'a' <= char <= 'p':
        count = ord(char) - ord('a') + 1
        is_lowercase = True
    elif 'A' <= char <= 'P':
        count = ord(char) - ord('A') + 1
        is_lowercase = False
    else:
        raise ValueError(f"Invalid position character: {char}")

    if turn == 1:
        # O's perspective: lowercase=X, uppercase=O
        return count if is_lowercase else -count
    else:
        # X's perspective: lowercase=O, uppercase=X
        return -count if is_lowercase else count


def encode_xgid(
    position: Position,
    cube_value: int = 1,
    cube_owner: CubeState = CubeState.CENTERED,
    dice: Optional[Tuple[int, int]] = None,
    on_roll: Player = Player.O,
    score_x: int = 0,
    score_o: int = 0,
    match_length: int = 0,
    crawford_jacoby: int = 0,
    max_cube: int = 256,
) -> str:
    """
    Encode a position and metadata as an XGID string.

    Args:
        position: The position to encode
        cube_value: Doubling cube value
        cube_owner: Who owns the cube
        dice: Dice values (if any)
        on_roll: Player on roll
        score_x: TOP player's score
        score_o: BOTTOM player's score
        match_length: Match length (0 for money)
        crawford_jacoby: Crawford/Jacoby setting
        max_cube: Maximum cube value

    Returns:
        XGID string
    """
    # Turn: 1 = BOTTOM (O), -1 = TOP (X)
    turn = 1 if on_roll == Player.O else -1

    # Encode position (turn-dependent)
    pos_str = _encode_position_string(position, turn)

    # Cube value as log2
    cube_value_log = 0
    temp_cube = cube_value
    while temp_cube > 1:
        temp_cube //= 2
        cube_value_log += 1

    # Cube position is perspective-dependent (flipped when X is on roll)
    if cube_owner == CubeState.CENTERED:
        cube_position = 0
    elif turn == 1:
        cube_position = -1 if cube_owner == CubeState.X_OWNS else 1
    else:
        cube_position = 1 if cube_owner == CubeState.X_OWNS else -1

    # Dice
    if dice:
        dice_str = f"{dice[0]}{dice[1]}"
    else:
        dice_str = "00"

    # Max cube as log2
    max_cube_log = 0
    temp = max_cube
    while temp > 1:
        temp //= 2
        max_cube_log += 1

    # Build XGID
    xgid = (
        f"XGID={pos_str}:"
        f"{cube_value_log}:{cube_position}:{turn}:{dice_str}:"
        f"{score_o}:{score_x}:"
        f"{crawford_jacoby}:{match_length}:{max_cube_log}"
    )

    return xgid


def _encode_position_string(position: Position, turn: int) -> str:
    """
    Encode a position to the 26-character XGID format.

    The encoding depends on whose turn it is:

    When turn=1 (O on roll):
    - Char 0: X's bar (points[0])
    - Chars 1-24: points in standard order (points[1-24])
    - Char 25: O's bar (points[25])

    When turn=-1 (X on roll):
    - Char 0: O's bar (points[25])
    - Chars 1-24: points in reversed order
    - Char 25: X's bar (points[0])

    Args:
        position: The position to encode
        turn: 1 if O on roll, -1 if X on roll

    Returns:
        26-character position string
    """
    chars = [''] * 26

    if turn == 1:
        # O on roll: standard perspective
        chars[0] = _encode_checker_count(position.points[0], turn)
        chars[25] = _encode_checker_count(position.points[25], turn)

        for i in range(1, 25):
            chars[i] = _encode_checker_count(position.points[i], turn)
    else:
        # X on roll: flipped perspective - bars and points are reversed
        chars[0] = _encode_checker_count(position.points[25], turn)  # O's bar
        chars[25] = _encode_checker_count(position.points[0], turn)  # X's bar

        # Board points: reverse mapping
        for i in range(1, 25):
            chars[25 - i] = _encode_checker_count(position.points[i], turn)

    return ''.join(chars)


def _encode_checker_count(count: int, turn: int) -> str:
    """
    Encode checker count to a single character.

    The uppercase/lowercase mapping depends on whose turn it is:

    When turn=1 (O on roll):
    - 0 = '-'
    - positive (X) = lowercase 'a' to 'p'
    - negative (O) = uppercase 'A' to 'P'

    When turn=-1 (X on roll):
    - 0 = '-'
    - positive (X) = uppercase 'A' to 'P'
    - negative (O) = lowercase 'a' to 'p'

    Args:
        count: Checker count (positive for X, negative for O, 0 for empty)
        turn: 1 if O on roll, -1 if X on roll

    Returns:
        Single character encoding
    """
    if count == 0:
        return '-'

    abs_count = abs(count)
    if abs_count > 16:
        abs_count = 16

    if turn == 1:
        # O's perspective: lowercase=X, uppercase=O
        if count > 0:
            return chr(ord('a') + abs_count - 1)
        else:
            return chr(ord('A') + abs_count - 1)
    else:
        # X's perspective: uppercase=X, lowercase=O
        if count > 0:
            return chr(ord('A') + abs_count - 1)
        else:
            return chr(ord('a') + abs_count - 1)



"""
Third-party code with different licenses.

This directory contains code from external projects that use licenses
other than the main project license (MIT). Each subdirectory has its
own LICENSE file.
"""



"""
Settings configuration dialog.
"""

import os
import sys
import subprocess
import tempfile
from pathlib import Path
from typing import Optional
import qtawesome as qta
from PySide6.QtWidgets import (
    QDialog, QVBoxLayout, QHBoxLayout, QFormLayout,
    QComboBox, QCheckBox, QLineEdit, QPushButton,
    QGroupBox, QFileDialog, QLabel, QDialogButtonBox
)
from PySide6.QtCore import Qt, Signal, QThread

from ankigammon.settings import Settings
from ankigammon.renderer.color_schemes import list_schemes


class GnuBGValidationWorker(QThread):
    """Worker thread for validating GnuBG executable without blocking UI."""

    # Signals to communicate with main thread
    validation_complete = Signal(str, str)  # (status_text, status_type)

    def __init__(self, gnubg_path: str):
        super().__init__()
        self.gnubg_path = gnubg_path

    def run(self):
        """Run validation in background thread."""
        path_obj = Path(self.gnubg_path)

        # Check if file exists
        if not path_obj.exists():
            self.validation_complete.emit("File not found", "error")
            return

        if not path_obj.is_file():
            self.validation_complete.emit("Not a file", "error")
            return

        # Create a simple command file (same approach as gnubg_analyzer)
        command_file = None
        try:
            # Create temp command file
            fd, command_file = tempfile.mkstemp(suffix=".txt", prefix="gnubg_test_")
            try:
                with os.fdopen(fd, 'w') as f:
                    # Simple command that should work on any gnubg
                    f.write("quit\n")
            except:
                os.close(fd)
                raise

            # Try to run gnubg with -t (text mode) and -c (command file)
            # Suppress console window on Windows; allow extra time for neural network loading
            kwargs = {
                'capture_output': True,
                'text': True,
                'timeout': 15
            }
            if sys.platform == 'win32':
                kwargs['creationflags'] = subprocess.CREATE_NO_WINDOW

            result = subprocess.run(
                [str(self.gnubg_path), "-t", "-c", command_file],
                **kwargs
            )

            # Check if it's actually GNU Backgammon
            output = result.stdout + result.stderr
            if "GNU Backgammon" in output or result.returncode == 0:
                # Check for GUI version and recommend CLI version on Windows
                exe_name = path_obj.stem.lower()
                if sys.platform == 'win32' and "cli" not in exe_name and exe_name == "gnubg":
                    self.validation_complete.emit(
                        "GUI version detected (use gnubg-cli.exe)",
                        "warning"
                    )
                else:
                    self.validation_complete.emit(
                        "Valid GnuBG executable",
                        "valid"
                    )
            else:
                self.validation_complete.emit("Not GNU Backgammon", "warning")

        except subprocess.TimeoutExpired:
            self.validation_complete.emit("Validation timeout", "warning")
        except Exception as e:
            self.validation_complete.emit(
                f"Cannot execute: {type(e).__name__}",
                "warning"
            )
        finally:
            # Clean up temp file
            if command_file:
                try:
                    os.unlink(command_file)
                except OSError:
                    pass


class SettingsDialog(QDialog):
    """
    Dialog for configuring application settings.

    Signals:
        settings_changed(Settings): Emitted when user saves changes
    """

    settings_changed = Signal(Settings)

    def __init__(self, settings: Settings, parent: Optional[QDialog] = None):
        super().__init__(parent)
        self.settings = settings
        self.original_settings = Settings()
        self.original_settings.color_scheme = settings.color_scheme
        self.original_settings.deck_name = settings.deck_name
        self.original_settings.show_options = settings.show_options
        self.original_settings.interactive_moves = settings.interactive_moves
        self.original_settings.export_method = settings.export_method
        self.original_settings.board_orientation = settings.board_orientation
        self.original_settings.gnubg_path = settings.gnubg_path
        self.original_settings.gnubg_analysis_ply = settings.gnubg_analysis_ply
        self.original_settings.generate_score_matrix = settings.generate_score_matrix
        self.original_settings.max_mcq_options = settings.max_mcq_options

        # Validation worker
        self.validation_worker: Optional[GnuBGValidationWorker] = None

        self.setWindowTitle("Settings")
        self.setModal(True)
        self.setMinimumWidth(500)

        self._setup_ui()
        self._load_settings()

    def _setup_ui(self):
        """Initialize the user interface."""
        layout = QVBoxLayout(self)

        # Anki settings group
        anki_group = self._create_anki_group()
        layout.addWidget(anki_group)

        # Card settings group
        card_group = self._create_card_group()
        layout.addWidget(card_group)

        # GnuBG settings group
        gnubg_group = self._create_gnubg_group()
        layout.addWidget(gnubg_group)

        # Dialog buttons
        button_box = QDialogButtonBox(
            QDialogButtonBox.Ok | QDialogButtonBox.Cancel
        )
        button_box.accepted.connect(self.accept)
        button_box.rejected.connect(self.reject)

        # Add cursor pointers to OK and Cancel buttons
        for button in button_box.buttons():
            button.setCursor(Qt.PointingHandCursor)

        layout.addWidget(button_box)

    def _create_anki_group(self) -> QGroupBox:
        """Create Anki settings group."""
        group = QGroupBox("Anki Export")
        form = QFormLayout(group)

        # Deck name
        self.txt_deck_name = QLineEdit()
        form.addRow("Default Deck Name:", self.txt_deck_name)

        # Export method
        self.cmb_export_method = QComboBox()
        self.cmb_export_method.addItems(["AnkiConnect", "APKG File"])
        self.cmb_export_method.setCursor(Qt.PointingHandCursor)
        form.addRow("Default Export Method:", self.cmb_export_method)

        return group

    def _create_card_group(self) -> QGroupBox:
        """Create card settings group."""
        group = QGroupBox("Card Appearance")
        form = QFormLayout(group)

        # Board theme
        self.cmb_color_scheme = QComboBox()
        self.cmb_color_scheme.addItems(list_schemes())
        self.cmb_color_scheme.setCursor(Qt.PointingHandCursor)
        form.addRow("Board Theme:", self.cmb_color_scheme)

        # Board orientation
        self.cmb_board_orientation = QComboBox()
        self.cmb_board_orientation.addItem("Counter-clockwise", "counter-clockwise")
        self.cmb_board_orientation.addItem("Clockwise", "clockwise")
        self.cmb_board_orientation.setCursor(Qt.PointingHandCursor)
        form.addRow("Board Orientation:", self.cmb_board_orientation)

        # Show options with max options dropdown on same line
        show_options_layout = QHBoxLayout()
        self.chk_show_options = QCheckBox("Show multiple choice options on card front")
        self.chk_show_options.setCursor(Qt.PointingHandCursor)
        show_options_layout.addWidget(self.chk_show_options)

        # Push max options to the right
        show_options_layout.addStretch()

        # Max options dropdown (on same line, right-aligned)
        self.lbl_max_options = QLabel("Max Options:")
        show_options_layout.addWidget(self.lbl_max_options)

        self.cmb_max_mcq_options = QComboBox()
        self.cmb_max_mcq_options.addItems([str(i) for i in range(2, 11)])
        self.cmb_max_mcq_options.setCursor(Qt.PointingHandCursor)
        self.cmb_max_mcq_options.setMaximumWidth(80)
        show_options_layout.addWidget(self.cmb_max_mcq_options)

        form.addRow(show_options_layout)

        # Connect checkbox to enable/disable dropdown
        self.chk_show_options.toggled.connect(self._on_show_options_toggled)

        # Interactive moves
        self.chk_interactive_moves = QCheckBox("Enable interactive move visualization")
        self.chk_interactive_moves.setCursor(Qt.PointingHandCursor)
        form.addRow(self.chk_interactive_moves)

        return group

    def _create_gnubg_group(self) -> QGroupBox:
        """Create GnuBG settings group."""
        group = QGroupBox("GnuBG Integration (Optional)")
        form = QFormLayout(group)

        # GnuBG path
        path_layout = QHBoxLayout()
        self.txt_gnubg_path = QLineEdit()
        btn_browse = QPushButton("Browse...")
        btn_browse.setCursor(Qt.PointingHandCursor)
        btn_browse.clicked.connect(self._browse_gnubg)
        path_layout.addWidget(self.txt_gnubg_path)
        path_layout.addWidget(btn_browse)
        form.addRow("GnuBG CLI Path:", path_layout)

        # Analysis depth
        self.cmb_gnubg_ply = QComboBox()
        self.cmb_gnubg_ply.addItems(["0", "1", "2", "3", "4"])
        self.cmb_gnubg_ply.setCursor(Qt.PointingHandCursor)
        form.addRow("Analysis Depth (ply):", self.cmb_gnubg_ply)

        # Score matrix generation
        matrix_layout = QHBoxLayout()
        self.chk_generate_score_matrix = QCheckBox("Generate score matrix for cube decisions")
        self.chk_generate_score_matrix.setCursor(Qt.PointingHandCursor)
        matrix_layout.addWidget(self.chk_generate_score_matrix)
        matrix_warning = QLabel("(time-consuming)")
        matrix_warning.setStyleSheet("font-size: 11px; color: #a6adc8; margin-left: 8px;")
        matrix_layout.addWidget(matrix_warning)
        matrix_layout.addStretch()
        form.addRow(matrix_layout)

        # Status display (icon + text in horizontal layout)
        status_layout = QHBoxLayout()
        self.lbl_gnubg_status_icon = QLabel()
        self.lbl_gnubg_status_text = QLabel()
        status_layout.addWidget(self.lbl_gnubg_status_icon)
        status_layout.addWidget(self.lbl_gnubg_status_text)
        status_layout.addStretch()
        form.addRow("Status:", status_layout)

        return group

    def _load_settings(self):
        """Load current settings into widgets."""
        self.txt_deck_name.setText(self.settings.deck_name)

        # Export method
        method_index = 0 if self.settings.export_method == "ankiconnect" else 1
        self.cmb_export_method.setCurrentIndex(method_index)

        # Color scheme
        scheme_index = list_schemes().index(self.settings.color_scheme)
        self.cmb_color_scheme.setCurrentIndex(scheme_index)

        # Board orientation
        orientation_index = 0 if self.settings.board_orientation == "counter-clockwise" else 1
        self.cmb_board_orientation.setCurrentIndex(orientation_index)

        self.chk_show_options.setChecked(self.settings.show_options)
        self.chk_interactive_moves.setChecked(self.settings.interactive_moves)

        # Max MCQ options dropdown (index is value minus 2)
        self.cmb_max_mcq_options.setCurrentIndex(self.settings.max_mcq_options - 2)

        # Initialize max options enabled state based on show options checkbox
        self._on_show_options_toggled(self.settings.show_options)

        # GnuBG
        if self.settings.gnubg_path:
            self.txt_gnubg_path.setText(self.settings.gnubg_path)
        self.cmb_gnubg_ply.setCurrentIndex(self.settings.gnubg_analysis_ply)
        self.chk_generate_score_matrix.setChecked(self.settings.generate_score_matrix)
        self._update_gnubg_status()

    def _browse_gnubg(self):
        """Browse for GnuBG executable."""
        # Platform-specific file filter
        if sys.platform == 'win32':
            file_filter = "Executables (*.exe);;All Files (*)"
        else:
            file_filter = "All Files (*)"

        file_path, _ = QFileDialog.getOpenFileName(
            self,
            "Select GnuBG Executable",
            "",
            file_filter
        )
        if file_path:
            self.txt_gnubg_path.setText(file_path)
            self._update_gnubg_status()

    def _update_gnubg_status(self):
        """Update GnuBG status label asynchronously."""
        # Cancel any running validation
        if self.validation_worker and self.validation_worker.isRunning():
            self.validation_worker.quit()
            self.validation_worker.wait()

        path = self.txt_gnubg_path.text()
        if not path:
            self.lbl_gnubg_status_icon.setPixmap(qta.icon('fa6s.circle', color='#6c7086').pixmap(18, 18))
            self.lbl_gnubg_status_text.setText("Not configured")
            self.lbl_gnubg_status_text.setStyleSheet("")
            return

        # Show loading state
        self.lbl_gnubg_status_icon.setPixmap(qta.icon('fa6s.spinner', color='#6c7086').pixmap(18, 18))
        self.lbl_gnubg_status_text.setText("Validating...")
        self.lbl_gnubg_status_text.setStyleSheet("color: gray;")

        # Start validation in background thread
        self.validation_worker = GnuBGValidationWorker(path)
        self.validation_worker.validation_complete.connect(self._on_validation_complete)
        self.validation_worker.start()

    def _on_validation_complete(self, status_text: str, status_type: str):
        """Handle validation completion."""
        # Determine icon based on status type
        if status_type == "valid":
            icon = qta.icon('fa6s.circle-check', color='#a6e3a1')
        elif status_type == "warning":
            icon = qta.icon('fa6s.triangle-exclamation', color='#fab387')
        elif status_type == "error":
            icon = qta.icon('fa6s.circle-xmark', color='#f38ba8')
        else:
            icon = None

        # Set icon and text separately
        if icon:
            self.lbl_gnubg_status_icon.setPixmap(icon.pixmap(18, 18))
        self.lbl_gnubg_status_text.setText(status_text)
        self.lbl_gnubg_status_text.setStyleSheet("")

    def _on_show_options_toggled(self, checked: bool):
        """Enable/disable max options dropdown based on show options checkbox."""
        self.lbl_max_options.setEnabled(checked)
        self.cmb_max_mcq_options.setEnabled(checked)

        # Add visual feedback for disabled state
        if checked:
            self.lbl_max_options.setStyleSheet("")
        else:
            self.lbl_max_options.setStyleSheet("color: #6c7086;")

    def accept(self):
        """Save settings and close dialog."""
        # Update settings object
        self.settings.deck_name = self.txt_deck_name.text()
        self.settings.export_method = (
            "ankiconnect" if self.cmb_export_method.currentIndex() == 0 else "apkg"
        )
        self.settings.color_scheme = self.cmb_color_scheme.currentText()
        self.settings.board_orientation = self.cmb_board_orientation.currentData()
        self.settings.show_options = self.chk_show_options.isChecked()
        self.settings.interactive_moves = self.chk_interactive_moves.isChecked()
        self.settings.max_mcq_options = self.cmb_max_mcq_options.currentIndex() + 2
        self.settings.gnubg_path = self.txt_gnubg_path.text() or None
        self.settings.gnubg_analysis_ply = self.cmb_gnubg_ply.currentIndex()
        self.settings.generate_score_matrix = self.chk_generate_score_matrix.isChecked()

        # Emit signal
        self.settings_changed.emit(self.settings)

        super().accept()

    def reject(self):
        """Restore original settings and close dialog."""
        # Clean up validation worker
        if self.validation_worker and self.validation_worker.isRunning():
            self.validation_worker.quit()
            self.validation_worker.wait()
        # Don't modify settings object
        super().reject()

    def closeEvent(self, event):
        """Clean up when dialog is closed."""
        # Clean up validation worker
        if self.validation_worker and self.validation_worker.isRunning():
            self.validation_worker.quit()
            self.validation_worker.wait()
        super().closeEvent(event)
